@@ -210,7 +210,7 @@ public class FakeTsiHandshaker implements TsiHandshaker {
 
   @Override
   public Object extractPeerObject() {
-    return AltsAuthContext.getDefaultInstance();
+    return AltsInternalContext.getDefaultInstance();
   }
 
   @Override
@@ -224,7 +224,7 @@ public class FakeTsiHandshaker implements TsiHandshaker {
 
   @Override
   public TsiFrameProtector createFrameProtector(ByteBufAllocator alloc) {
-    return createFrameProtector(AltsTsiFrameProtector.getMaxAllowedFrameBytes(), alloc);
+    return createFrameProtector(AltsTsiFrameProtector.getMinFrameSize(), alloc);
   }
 
   @Override
