@@ -55,8 +55,11 @@ public abstract class ManagedChannelBuilder<T extends ManagedChannelBuilder<T>>
    * </ul>
    *
    * <p>An authority string will be converted to a {@code NameResolver}-compliant URI, which has
-   * {@code "dns"} as the scheme, no authority, and the original authority string as its path after
-   * properly escaped. Example authority strings:
+   * the scheme from the name resolver with the highest priority (e.g. {@code "dns"}),
+   * no authority, and the original authority string as its path after properly escaped.
+   * We recommend libraries to specify the schema explicitly if it is known, since libraries cannot
+   * know which NameResolver will be default during runtime.
+   * Example authority strings:
    * <ul>
    *   <li>{@code "localhost"}</li>
    *   <li>{@code "127.0.0.1"}</li>
@@ -121,12 +124,6 @@ public abstract class ManagedChannelBuilder<T extends ManagedChannelBuilder<T>>
     throw new UnsupportedOperationException();
   }
 
-  @Deprecated
-  @ExperimentalApi("https://github.com/grpc/grpc-java/issues/6279")
-  public T blockingExecutor(Executor executor) {
-    return offloadExecutor(executor);
-  }
-
   /**
    * Adds interceptors that will be called before the channel performs its real work. This is
    * functionally equivalent to using {@link ClientInterceptors#intercept(Channel, List)}, but while
@@ -166,6 +163,9 @@ public abstract class ManagedChannelBuilder<T extends ManagedChannelBuilder<T>>
    * Overrides the authority used with TLS and HTTP virtual hosting. It does not change what host is
    * actually connected to. Is commonly in the form {@code host:port}.
    *
+   * <p>If the channel builder overrides authority, any authority override from name resolution
+   * result (via {@link EquivalentAddressGroup#ATTR_AUTHORITY_OVERRIDE}) will be discarded.
+   *
    * <p>This method is intended for testing, but may safely be used outside of tests as an
    * alternative to DNS overrides.
    *
@@ -185,6 +185,7 @@ public abstract class ManagedChannelBuilder<T extends ManagedChannelBuilder<T>>
    * not perform HTTP/1.1 upgrades.
    *
    * @return this
+   * @throws IllegalStateException if ChannelCredentials were provided when constructing the builder
    * @throws UnsupportedOperationException if plaintext mode is not supported.
    * @since 1.11.0
    */
@@ -196,6 +197,7 @@ public abstract class ManagedChannelBuilder<T extends ManagedChannelBuilder<T>>
    * Makes the client use TLS.
    *
    * @return this
+   * @throws IllegalStateException if ChannelCredentials were provided when constructing the builder
    * @throws UnsupportedOperationException if transport security is not supported.
    * @since 1.9.0
    */
@@ -206,8 +208,8 @@ public abstract class ManagedChannelBuilder<T extends ManagedChannelBuilder<T>>
 
   /**
    * Provides a custom {@link NameResolver.Factory} for the channel. If this method is not called,
-   * the builder will try the providers listed by {@link NameResolverProvider#providers()} for the
-   * given target.
+   * the builder will try the providers registered in the default {@link NameResolverRegistry} for
+   * the given target.
    *
    * <p>This method should rarely be used, as name resolvers should provide a {@code
    * NameResolverProvider} and users rely on service loading to find implementations in the class
@@ -216,7 +218,12 @@ public abstract class ManagedChannelBuilder<T extends ManagedChannelBuilder<T>>
    *
    * @return this
    * @since 1.0.0
+   * @deprecated Most usages should use a globally-registered {@link NameResolverProvider} instead,
+   *     with either the SPI mechanism or {@link NameResolverRegistry#register}. Replacements for
+   *     all use-cases are not necessarily available yet. See
+   *     <a href="https://github.com/grpc/grpc-java/issues/7133">#7133</a>.
    */
+  @Deprecated
   @ExperimentalApi("https://github.com/grpc/grpc-java/issues/1770")
   public abstract T nameResolverFactory(NameResolver.Factory resolverFactory);
 
@@ -382,9 +389,10 @@ public abstract class ManagedChannelBuilder<T extends ManagedChannelBuilder<T>>
   }
 
   /**
-   * Sets max number of retry attempts. The total number of retry attempts for each RPC will not
-   * exceed this number even if service config may allow a higher number. Setting this number to
-   * zero is not effectively the same as {@code disableRetry()} because the former does not disable
+   * Sets the maximum number of retry attempts that may be configured by the service config. If the
+   * service config specifies a larger value it will be reduced to this value.  Setting this number
+   * to zero is not effectively the same as {@code disableRetry()} because the former does not
+   * disable
    * <a
    * href="https://github.com/grpc/proposal/blob/master/A6-client-retries.md#transparent-retries">
    * transparent retry</a>.
@@ -401,8 +409,8 @@ public abstract class ManagedChannelBuilder<T extends ManagedChannelBuilder<T>>
   }
 
   /**
-   * Sets max number of hedged attempts. The total number of hedged attempts for each RPC will not
-   * exceed this number even if service config may allow a higher number.
+   * Sets the maximum number of hedged attempts that may be configured by the service config. If the
+   * service config specifies a larger value it will be reduced to this value.
    *
    * <p>This method may not work as expected for the current release because retry is not fully
    * implemented yet.
@@ -452,28 +460,27 @@ public abstract class ManagedChannelBuilder<T extends ManagedChannelBuilder<T>>
 
 
   /**
-   * Disables the retry and hedging mechanism provided by the gRPC library. This is designed for the
+   * Disables the retry and hedging subsystem provided by the gRPC library. This is designed for the
    * case when users have their own retry implementation and want to avoid their own retry taking
    * place simultaneously with the gRPC library layer retry.
    *
    * @return this
    * @since 1.11.0
    */
-  @ExperimentalApi("https://github.com/grpc/grpc-java/issues/3982")
   public T disableRetry() {
     throw new UnsupportedOperationException();
   }
 
   /**
-   * Enables the retry and hedging mechanism provided by the gRPC library.
-   *
-   * <p>For the current release, this method may have a side effect that disables Census stats and
-   * tracing.
+   * Enables the retry and hedging subsystem which will use
+   * <a href="https://github.com/grpc/proposal/blob/master/A6-client-retries.md#integration-with-service-config">
+   * per-method configuration</a>. If a method is unconfigured, it will be limited to
+   * transparent retries, which are safe for non-idempotent RPCs. Service config is ideally provided
+   * by the name resolver, but may also be specified via {@link #defaultServiceConfig}.
    *
    * @return this
    * @since 1.11.0
    */
-  @ExperimentalApi("https://github.com/grpc/grpc-java/issues/3982")
   public T enableRetry() {
     throw new UnsupportedOperationException();
   }
