@@ -329,7 +329,7 @@ public class ClientInterceptorsTest {
     assertSame(listener, call.listener);
     assertSame(headers, call.headers);
     interceptedCall.sendMessage(null /*request*/);
-    assertThat(call.messages).containsExactly((Void) null /*request*/);
+    assertThat(call.messages).containsExactly((String) null);
     interceptedCall.halfClose();
     assertTrue(call.halfClosed);
     interceptedCall.request(1);
