@@ -10,35 +10,37 @@ BASE_DIR="$(pwd)"
 
 cd "$BASE_DIR/github/grpc-java"
 
-export GRADLE_OPTS=-Xmx512m
 export LDFLAGS=-L/tmp/protobuf/lib
 export CXXFLAGS=-I/tmp/protobuf/include
 export LD_LIBRARY_PATH=/tmp/protobuf/lib
 export OS_NAME=$(uname)
 
+cat <<EOF >> gradle.properties
+# defaults to -Xmx512m -XX:MaxMetaspaceSize=256m
+# https://docs.gradle.org/current/userguide/build_environment.html#sec:configuring_jvm_memory
+# Increased due to java.lang.OutOfMemoryError: Metaspace failures, "JVM heap
+# space is exhausted", and to increase build speed
+org.gradle.jvmargs=-Xmx2048m -XX:MaxMetaspaceSize=512m
+EOF
+
 echo y | ${ANDROID_HOME}/tools/bin/sdkmanager "build-tools;28.0.3"
 
 # Proto deps
 buildscripts/make_dependencies.sh
 
-./gradlew publishToMavenLocal
+./gradlew \
+    :grpc-android-interop-testing:build \
+    :grpc-android:build \
+    :grpc-cronet:build \
+    :grpc-binder:build \
+    assembleAndroidTest \
+    publishToMavenLocal
 
-# Build grpc-cronet
-
-pushd cronet
-../gradlew build
-popd
-
-# Build grpc-android
-
-pushd android
-../gradlew build
-popd
-
-# Build android-interop-testing
-pushd android-interop-testing
-../gradlew build
-popd
+if [[ ! -z $(git status --porcelain) ]]; then
+  git status
+  echo "Error Working directory is not clean. Forget to commit generated files?"
+  exit 1
+fi
 
 # Build examples
 
@@ -48,9 +50,8 @@ cd ../routeguide
 ../../gradlew build
 cd ../helloworld
 ../../gradlew build
-
-cd "$BASE_DIR/github/grpc-java/examples/example-kotlin/android/helloworld/"
-../../../gradlew build
+cd ../strictmode
+../../gradlew build
 
 # Skip APK size and dex count comparisons for non-PR builds
 
@@ -85,8 +86,9 @@ new_apk_size="$(stat --printf=%s $HELLO_WORLD_OUTPUT_DIR/apk/release/app-release
 # Get the APK size and dex count stats using the pull request base commit
 
 cd $BASE_DIR/github/grpc-java
-git checkout HEAD^
 ./gradlew clean
+git checkout HEAD^
+./gradlew --stop  # use a new daemon to build the previous commit
 ./gradlew publishToMavenLocal
 cd examples/android/helloworld/
 ../../gradlew build
