@@ -36,6 +36,7 @@ GRADLE_FLAGS+=" -PtargetArch=$ARCH"
 GRADLE_FLAGS+=" -Pcheckstyle.ignoreFailures=false"
 GRADLE_FLAGS+=" -PfailOnWarnings=true"
 GRADLE_FLAGS+=" -PerrorProne=true"
+GRADLE_FLAGS+=" -PskipAndroid=true"
 GRADLE_FLAGS+=" -Dorg.gradle.parallel=true"
 export GRADLE_OPTS="-Xmx512m"
 
@@ -56,15 +57,30 @@ if [[ -z "${SKIP_TESTS:-}" ]]; then
     exit 1
   fi
   # Run tests
-  ./gradlew build $GRADLE_FLAGS
+  ./gradlew build :grpc-all:jacocoTestReport $GRADLE_FLAGS
   pushd examples
   ./gradlew clean $GRADLE_FLAGS
   ./gradlew build $GRADLE_FLAGS
   # --batch-mode reduces log spam
-  mvn clean verify --batch-mode
+  mvn verify --batch-mode
+  popd
+  pushd examples/example-alts
+  ../gradlew build $GRADLE_FLAGS
+  popd
+  pushd examples/example-hostname
+  ../gradlew build $GRADLE_FLAGS
+  mvn verify --batch-mode
   popd
   pushd examples/example-tls
-  mvn clean verify --batch-mode
+  ../gradlew build $GRADLE_FLAGS
+  mvn verify --batch-mode
+  popd
+  pushd examples/example-jwt-auth
+  ../gradlew build $GRADLE_FLAGS
+  mvn verify --batch-mode
+  popd
+  pushd examples/example-xds
+  ../gradlew build $GRADLE_FLAGS
   popd
   # TODO(zpencer): also build the GAE examples
 fi
@@ -80,8 +96,16 @@ if [[ -z "${ALL_ARTIFACTS:-}" ]]; then
   ./gradlew grpc-compiler:build grpc-compiler:publish $GRADLE_FLAGS \
     -Dorg.gradle.parallel=false -PrepositoryDir=$LOCAL_MVN_TEMP
 else
-  ./gradlew publish $GRADLE_FLAGS \
+  ./gradlew publish :grpc-core:versionFile $GRADLE_FLAGS \
     -Dorg.gradle.parallel=false -PrepositoryDir=$LOCAL_MVN_TEMP
+  pushd examples/example-hostname
+  ../gradlew jibBuildTar $GRADLE_FLAGS
+  popd
+
+  readonly OTHER_ARTIFACT_DIR="${OTHER_ARTIFACT_DIR:-$GRPC_JAVA_DIR/artifacts}"
+  mkdir -p "$OTHER_ARTIFACT_DIR"
+  cp core/build/version "$OTHER_ARTIFACT_DIR"/
+  cp examples/example-hostname/build/example-hostname.* "$OTHER_ARTIFACT_DIR"/
 fi
 
 readonly MVN_ARTIFACT_DIR="${MVN_ARTIFACT_DIR:-$GRPC_JAVA_DIR/mvn-artifacts}"
