@@ -21,10 +21,12 @@ import static io.grpc.internal.GrpcUtil.TIMEOUT_KEY;
 import static java.lang.Math.max;
 
 import com.google.common.base.MoreObjects;
+import com.google.common.base.Optional;
 import com.google.common.util.concurrent.ListenableFuture;
 import com.google.common.util.concurrent.SettableFuture;
 import io.grpc.Attributes;
 import io.grpc.CallOptions;
+import io.grpc.ClientStreamTracer;
 import io.grpc.Compressor;
 import io.grpc.Deadline;
 import io.grpc.Decompressor;
@@ -40,6 +42,7 @@ import io.grpc.ServerStreamTracer;
 import io.grpc.Status;
 import io.grpc.internal.ClientStream;
 import io.grpc.internal.ClientStreamListener;
+import io.grpc.internal.ClientStreamListener.RpcProgress;
 import io.grpc.internal.ConnectionClientTransport;
 import io.grpc.internal.GrpcAttributes;
 import io.grpc.internal.GrpcUtil;
@@ -48,6 +51,7 @@ import io.grpc.internal.InsightBuilder;
 import io.grpc.internal.ManagedClientTransport;
 import io.grpc.internal.NoopClientStream;
 import io.grpc.internal.ObjectPool;
+import io.grpc.internal.ServerListener;
 import io.grpc.internal.ServerStream;
 import io.grpc.internal.ServerStreamListener;
 import io.grpc.internal.ServerTransport;
@@ -57,7 +61,8 @@ import io.grpc.internal.StreamListener;
 import java.io.InputStream;
 import java.util.ArrayDeque;
 import java.util.ArrayList;
-import java.util.HashSet;
+import java.util.Collections;
+import java.util.IdentityHashMap;
 import java.util.List;
 import java.util.Set;
 import java.util.concurrent.Executor;
@@ -79,7 +84,9 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
   private final int clientMaxInboundMetadataSize;
   private final String authority;
   private final String userAgent;
+  private final Optional<ServerListener> optionalServerListener;
   private int serverMaxInboundMetadataSize;
+  private final boolean includeCauseWithStatus;
   private ObjectPool<ScheduledExecutorService> serverSchedulerPool;
   private ScheduledExecutorService serverScheduler;
   private ServerTransportListener serverTransportListener;
@@ -92,7 +99,8 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
   @GuardedBy("this")
   private Status shutdownStatus;
   @GuardedBy("this")
-  private Set<InProcessStream> streams = new HashSet<>();
+  private final Set<InProcessStream> streams = Collections.newSetFromMap(
+          new IdentityHashMap<InProcessStream, Boolean>());
   @GuardedBy("this")
   private List<ServerStreamTracer.Factory> serverStreamTracerFactories;
   private final Attributes attributes;
@@ -111,9 +119,9 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
         }
       };
 
-  public InProcessTransport(
-      String name, int maxInboundMetadataSize, String authority, String userAgent,
-      Attributes eagAttrs) {
+  private InProcessTransport(String name, int maxInboundMetadataSize, String authority,
+      String userAgent, Attributes eagAttrs,
+      Optional<ServerListener> optionalServerListener, boolean includeCauseWithStatus) {
     this.name = name;
     this.clientMaxInboundMetadataSize = maxInboundMetadataSize;
     this.authority = authority;
@@ -125,21 +133,47 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
         .set(Grpc.TRANSPORT_ATTR_REMOTE_ADDR, new InProcessSocketAddress(name))
         .set(Grpc.TRANSPORT_ATTR_LOCAL_ADDR, new InProcessSocketAddress(name))
         .build();
+    this.optionalServerListener = optionalServerListener;
     logId = InternalLogId.allocate(getClass(), name);
+    this.includeCauseWithStatus = includeCauseWithStatus;
+  }
+
+  public InProcessTransport(
+      String name, int maxInboundMetadataSize, String authority, String userAgent,
+      Attributes eagAttrs, boolean includeCauseWithStatus) {
+    this(name, maxInboundMetadataSize, authority, userAgent, eagAttrs,
+        Optional.<ServerListener>absent(), includeCauseWithStatus);
+  }
+
+  InProcessTransport(
+      String name, int maxInboundMetadataSize, String authority, String userAgent,
+      Attributes eagAttrs, ObjectPool<ScheduledExecutorService> serverSchedulerPool,
+      List<ServerStreamTracer.Factory> serverStreamTracerFactories,
+      ServerListener serverListener) {
+    this(name, maxInboundMetadataSize, authority, userAgent, eagAttrs,
+        Optional.of(serverListener), false);
+    this.serverMaxInboundMetadataSize = maxInboundMetadataSize;
+    this.serverSchedulerPool = serverSchedulerPool;
+    this.serverStreamTracerFactories = serverStreamTracerFactories;
   }
 
   @CheckReturnValue
   @Override
   public synchronized Runnable start(ManagedClientTransport.Listener listener) {
     this.clientTransportListener = listener;
-    InProcessServer server = InProcessServer.findServer(name);
-    if (server != null) {
-      serverMaxInboundMetadataSize = server.getMaxInboundMetadataSize();
-      serverSchedulerPool = server.getScheduledExecutorServicePool();
+    if (optionalServerListener.isPresent()) {
       serverScheduler = serverSchedulerPool.getObject();
-      serverStreamTracerFactories = server.getStreamTracerFactories();
-      // Must be semi-initialized; past this point, can begin receiving requests
-      serverTransportListener = server.register(this);
+      serverTransportListener = optionalServerListener.get().transportCreated(this);
+    } else {
+      InProcessServer server = InProcessServer.findServer(name);
+      if (server != null) {
+        serverMaxInboundMetadataSize = server.getMaxInboundMetadataSize();
+        serverSchedulerPool = server.getScheduledExecutorServicePool();
+        serverScheduler = serverSchedulerPool.getObject();
+        serverStreamTracerFactories = server.getStreamTracerFactories();
+        // Must be semi-initialized; past this point, can begin receiving requests
+        serverTransportListener = server.register(this);
+      }
     }
     if (serverTransportListener == null) {
       shutdownStatus = Status.UNAVAILABLE.withDescription("Could not find server: " + name);
@@ -172,10 +206,12 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
 
   @Override
   public synchronized ClientStream newStream(
-      final MethodDescriptor<?, ?> method, final Metadata headers, final CallOptions callOptions) {
+      MethodDescriptor<?, ?> method, Metadata headers, CallOptions callOptions,
+      ClientStreamTracer[] tracers) {
+    StatsTraceContext statsTraceContext =
+        StatsTraceContext.newClientContext(tracers, getAttributes(), headers);
     if (shutdownStatus != null) {
-      return failedClientStream(
-          StatsTraceContext.newClientContext(callOptions, attributes, headers), shutdownStatus);
+      return failedClientStream(statsTraceContext, shutdownStatus);
     }
 
     headers.put(GrpcUtil.USER_AGENT_KEY, userAgent);
@@ -193,12 +229,12 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
                 "Request metadata larger than %d: %d",
                 serverMaxInboundMetadataSize,
                 metadataSize));
-        return failedClientStream(
-            StatsTraceContext.newClientContext(callOptions, attributes, headers), status);
+        return failedClientStream(statsTraceContext, status);
       }
     }
 
-    return new InProcessStream(method, headers, callOptions, authority).clientStream;
+    return new InProcessStream(method, headers, callOptions, authority, statsTraceContext)
+        .clientStream;
   }
 
   private ClientStream failedClientStream(
@@ -208,7 +244,7 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
         public void start(ClientStreamListener listener) {
           statsTraceCtx.clientOutboundHeaders();
           statsTraceCtx.streamClosed(status);
-          listener.closed(status, new Metadata());
+          listener.closed(status, RpcProgress.PROCESSED, new Metadata());
         }
       };
   }
@@ -344,12 +380,12 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
 
     private InProcessStream(
         MethodDescriptor<?, ?> method, Metadata headers, CallOptions callOptions,
-        String authority) {
+        String authority , StatsTraceContext statsTraceContext) {
       this.method = checkNotNull(method, "method");
       this.headers = checkNotNull(headers, "headers");
       this.callOptions = checkNotNull(callOptions, "callOptions");
       this.authority = authority;
-      this.clientStream = new InProcessClientStream(callOptions, headers);
+      this.clientStream = new InProcessClientStream(callOptions, statsTraceContext);
       this.serverStream = new InProcessServerStream(method, headers);
     }
 
@@ -437,7 +473,8 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
           closed = true;
           clientStream.statsTraceCtx.clientInboundTrailers(clientNotifyTrailers);
           clientStream.statsTraceCtx.streamClosed(clientNotifyStatus);
-          clientStreamListener.closed(clientNotifyStatus, clientNotifyTrailers);
+          clientStreamListener.closed(
+              clientNotifyStatus, RpcProgress.PROCESSED, clientNotifyTrailers);
         }
         boolean nowReady = clientRequested > 0;
         return !previouslyReady && nowReady;
@@ -538,7 +575,7 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
 
       /** clientStream.serverClosed() must be called before this method */
       private void notifyClientClose(Status status, Metadata trailers) {
-        Status clientStatus = stripCause(status);
+        Status clientStatus = cleanStatus(status, includeCauseWithStatus);
         synchronized (this) {
           if (closed) {
             return;
@@ -547,7 +584,7 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
             closed = true;
             clientStream.statsTraceCtx.clientInboundTrailers(trailers);
             clientStream.statsTraceCtx.streamClosed(clientStatus);
-            clientStreamListener.closed(clientStatus, trailers);
+            clientStreamListener.closed(clientStatus, RpcProgress.PROCESSED, trailers);
           } else {
             clientNotifyStatus = clientStatus;
             clientNotifyTrailers = trailers;
@@ -583,7 +620,7 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
           }
         }
         clientStream.statsTraceCtx.streamClosed(clientStatus);
-        clientStreamListener.closed(clientStatus, new Metadata());
+        clientStreamListener.closed(clientStatus, RpcProgress.PROCESSED, new Metadata());
         return true;
       }
 
@@ -592,6 +629,9 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
         // noop
       }
 
+      @Override
+      public void optimizeForDirectExecutor() {}
+
       @Override
       public void setCompressor(Compressor compressor) {}
 
@@ -636,9 +676,10 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
       @GuardedBy("this")
       private int outboundSeqNo;
 
-      InProcessClientStream(CallOptions callOptions, Metadata headers) {
+      InProcessClientStream(
+          CallOptions callOptions, StatsTraceContext statsTraceContext) {
         this.callOptions = callOptions;
-        statsTraceCtx = StatsTraceContext.newClientContext(callOptions, attributes, headers);
+        statsTraceCtx = statsTraceContext;
       }
 
       private synchronized void setListener(ServerStreamListener listener) {
@@ -718,7 +759,7 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
       // Must be thread-safe for shutdownNow()
       @Override
       public void cancel(Status reason) {
-        Status serverStatus = stripCause(reason);
+        Status serverStatus = cleanStatus(reason, includeCauseWithStatus);
         if (!internalCancel(serverStatus, serverStatus)) {
           return;
         }
@@ -788,6 +829,9 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
         return attributes;
       }
 
+      @Override
+      public void optimizeForDirectExecutor() {}
+
       @Override
       public void setCompressor(Compressor compressor) {}
 
@@ -817,19 +861,25 @@ final class InProcessTransport implements ServerTransport, ConnectionClientTrans
   }
 
   /**
-   * Returns a new status with the same code and description, but stripped of any other information
-   * (i.e. cause).
+   * Returns a new status with the same code and description.
+   * If includeCauseWithStatus is true, cause is also included.
    *
-   * <p>This is, so that the InProcess transport behaves in the same way as the other transports,
-   * when exchanging statuses between client and server and vice versa.
+   * <p>For InProcess transport to behave in the same way as the other transports,
+   * when exchanging statuses between client and server and vice versa,
+   * the cause should be excluded from the status.
+   * For easier debugging, the status may be optionally included.
    */
-  private static Status stripCause(Status status) {
+  private static Status cleanStatus(Status status, boolean includeCauseWithStatus) {
     if (status == null) {
       return null;
     }
-    return Status
+    Status clientStatus = Status
         .fromCodeValue(status.getCode().value())
         .withDescription(status.getDescription());
+    if (includeCauseWithStatus) {
+      clientStatus = clientStatus.withCause(status.getCause());
+    }
+    return clientStatus;
   }
 
   private static class SingleMessageProducer implements StreamListener.MessageProducer {
