@@ -20,6 +20,7 @@ import static com.google.common.base.Preconditions.checkNotNull;
 
 import com.google.common.annotations.VisibleForTesting;
 import com.google.common.base.MoreObjects;
+import com.google.common.base.Objects;
 import com.google.common.base.Preconditions;
 import com.google.common.base.Stopwatch;
 import com.google.common.base.Throwables;
@@ -45,7 +46,6 @@ import java.util.Collections;
 import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
-import java.util.Map.Entry;
 import java.util.Random;
 import java.util.Set;
 import java.util.concurrent.Executor;
@@ -63,7 +63,7 @@ import javax.annotation.Nullable;
  *
  * @see DnsNameResolverProvider
  */
-final class DnsNameResolver extends NameResolver {
+public class DnsNameResolver extends NameResolver {
 
   private static final Logger logger = Logger.getLogger(DnsNameResolver.class.getName());
 
@@ -85,8 +85,6 @@ final class DnsNameResolver extends NameResolver {
 
   // From https://github.com/grpc/proposal/blob/master/A2-service-configs-in-dns.md
   private static final String SERVICE_CONFIG_NAME_PREFIX = "_grpc_config.";
-  // From https://github.com/grpc/proposal/blob/master/A5-grpclb-in-dns.md
-  private static final String GRPCLB_NAME_PREFIX = "_grpclb._tcp.";
 
   private static final String JNDI_PROPERTY =
       System.getProperty("io.grpc.internal.DnsNameResolverProvider.enable_jndi", "true");
@@ -115,7 +113,7 @@ final class DnsNameResolver extends NameResolver {
   @VisibleForTesting
   static boolean enableJndiLocalhost = Boolean.parseBoolean(JNDI_LOCALHOST_PROPERTY);
   @VisibleForTesting
-  static boolean enableTxt = Boolean.parseBoolean(JNDI_TXT_PROPERTY);
+  protected static boolean enableTxt = Boolean.parseBoolean(JNDI_TXT_PROPERTY);
 
   private static final ResourceResolverFactory resourceResolverFactory =
       getResourceResolverFactory(DnsNameResolver.class.getClassLoader());
@@ -128,7 +126,7 @@ final class DnsNameResolver extends NameResolver {
 
   private final Random random = new Random();
 
-  private volatile AddressResolver addressResolver = JdkAddressResolver.INSTANCE;
+  protected volatile AddressResolver addressResolver = JdkAddressResolver.INSTANCE;
   private final AtomicReference<ResourceResolver> resourceResolver = new AtomicReference<>();
 
   private final String authority;
@@ -142,13 +140,13 @@ final class DnsNameResolver extends NameResolver {
 
   // Following fields must be accessed from syncContext
   private final Stopwatch stopwatch;
-  private ResolutionResults cachedResolutionResults;
+  protected boolean resolved;
   private boolean shutdown;
   private Executor executor;
 
   /** True if using an executor resource that should be released after use. */
   private final boolean usingExecutorResource;
-  private final boolean enableSrv;
+  private final ServiceConfigParser serviceConfigParser;
 
   private boolean resolving;
 
@@ -156,15 +154,14 @@ final class DnsNameResolver extends NameResolver {
   // from any thread.
   private NameResolver.Listener2 listener;
 
-  DnsNameResolver(
+  protected DnsNameResolver(
       @Nullable String nsAuthority,
       String name,
       Args args,
       Resource<Executor> executorResource,
       Stopwatch stopwatch,
-      boolean isAndroid,
-      boolean enableSrv) {
-    Preconditions.checkNotNull(args, "args");
+      boolean isAndroid) {
+    checkNotNull(args, "args");
     // TODO: if a DNS server is provided as nsAuthority, use it.
     // https://www.captechconsulting.com/blogs/accessing-the-dusty-corners-of-dns-with-java
     this.executorResource = executorResource;
@@ -180,14 +177,13 @@ final class DnsNameResolver extends NameResolver {
     } else {
       port = nameUri.getPort();
     }
-    this.proxyDetector = Preconditions.checkNotNull(args.getProxyDetector(), "proxyDetector");
+    this.proxyDetector = checkNotNull(args.getProxyDetector(), "proxyDetector");
     this.cacheTtlNanos = getNetworkAddressCacheTtlNanos(isAndroid);
-    this.stopwatch = Preconditions.checkNotNull(stopwatch, "stopwatch");
-    this.syncContext =
-        Preconditions.checkNotNull(args.getSynchronizationContext(), "syncContext");
+    this.stopwatch = checkNotNull(stopwatch, "stopwatch");
+    this.syncContext = checkNotNull(args.getSynchronizationContext(), "syncContext");
     this.executor = args.getOffloadExecutor();
     this.usingExecutorResource = executor == null;
-    this.enableSrv = enableSrv;
+    this.serviceConfigParser = checkNotNull(args.getServiceConfigParser(), "serviceConfigParser");
   }
 
   @Override
@@ -195,13 +191,18 @@ final class DnsNameResolver extends NameResolver {
     return authority;
   }
 
+  @VisibleForTesting
+  protected String getHost() {
+    return host;
+  }
+
   @Override
   public void start(Listener2 listener) {
     Preconditions.checkState(this.listener == null, "already started");
     if (usingExecutorResource) {
       executor = SharedResourceHolder.get(executorResource);
     }
-    this.listener = Preconditions.checkNotNull(listener, "listener");
+    this.listener = checkNotNull(listener, "listener");
     resolve();
   }
 
@@ -211,11 +212,92 @@ final class DnsNameResolver extends NameResolver {
     resolve();
   }
 
+  private List<EquivalentAddressGroup> resolveAddresses() {
+    List<? extends InetAddress> addresses;
+    Exception addressesException = null;
+    try {
+      addresses = addressResolver.resolveAddress(host);
+    } catch (Exception e) {
+      addressesException = e;
+      Throwables.throwIfUnchecked(e);
+      throw new RuntimeException(e);
+    } finally {
+      if (addressesException != null) {
+        logger.log(Level.FINE, "Address resolution failure", addressesException);
+      }
+    }
+    // Each address forms an EAG
+    List<EquivalentAddressGroup> servers = new ArrayList<>(addresses.size());
+    for (InetAddress inetAddr : addresses) {
+      servers.add(new EquivalentAddressGroup(new InetSocketAddress(inetAddr, port)));
+    }
+    return Collections.unmodifiableList(servers);
+  }
+
+  @Nullable
+  private ConfigOrError resolveServiceConfig() {
+    List<String> txtRecords = Collections.emptyList();
+    ResourceResolver resourceResolver = getResourceResolver();
+    if (resourceResolver != null) {
+      try {
+        txtRecords = resourceResolver.resolveTxt(SERVICE_CONFIG_NAME_PREFIX + host);
+      } catch (Exception e) {
+        logger.log(Level.FINE, "ServiceConfig resolution failure", e);
+      }
+    }
+    if (!txtRecords.isEmpty()) {
+      ConfigOrError rawServiceConfig = parseServiceConfig(txtRecords, random, getLocalHostname());
+      if (rawServiceConfig != null) {
+        if (rawServiceConfig.getError() != null) {
+          return ConfigOrError.fromError(rawServiceConfig.getError());
+        }
+
+        @SuppressWarnings("unchecked")
+        Map<String, ?> verifiedRawServiceConfig = (Map<String, ?>) rawServiceConfig.getConfig();
+        return serviceConfigParser.parseServiceConfig(verifiedRawServiceConfig);
+      }
+    } else {
+      logger.log(Level.FINE, "No TXT records found for {0}", new Object[]{host});
+    }
+    return null;
+  }
+
+  @Nullable
+  private EquivalentAddressGroup detectProxy() throws IOException {
+    InetSocketAddress destination =
+        InetSocketAddress.createUnresolved(host, port);
+    ProxiedSocketAddress proxiedAddr = proxyDetector.proxyFor(destination);
+    if (proxiedAddr != null) {
+      return new EquivalentAddressGroup(proxiedAddr);
+    }
+    return null;
+  }
+
+  /**
+   * Main logic of name resolution.
+   */
+  protected InternalResolutionResult doResolve(boolean forceTxt) {
+    InternalResolutionResult result = new InternalResolutionResult();
+    try {
+      result.addresses = resolveAddresses();
+    } catch (Exception e) {
+      if (!forceTxt) {
+        result.error =
+            Status.UNAVAILABLE.withDescription("Unable to resolve host " + host).withCause(e);
+        return result;
+      }
+    }
+    if (enableTxt) {
+      result.config = resolveServiceConfig();
+    }
+    return result;
+  }
+
   private final class Resolve implements Runnable {
     private final Listener2 savedListener;
 
     Resolve(Listener2 savedListener) {
-      this.savedListener = Preconditions.checkNotNull(savedListener, "savedListener");
+      this.savedListener = checkNotNull(savedListener, "savedListener");
     }
 
     @Override
@@ -223,106 +305,50 @@ final class DnsNameResolver extends NameResolver {
       if (logger.isLoggable(Level.FINER)) {
         logger.finer("Attempting DNS resolution of " + host);
       }
+      InternalResolutionResult result = null;
       try {
-        resolveInternal();
-      } finally {
-        syncContext.execute(new Runnable() {
-            @Override
-            public void run() {
-              resolving = false;
-            }
-          });
-      }
-    }
-
-    @VisibleForTesting
-    void resolveInternal() {
-      InetSocketAddress destination =
-          InetSocketAddress.createUnresolved(host, port);
-      ProxiedSocketAddress proxiedAddr;
-      try {
-        proxiedAddr = proxyDetector.proxyFor(destination);
+        EquivalentAddressGroup proxiedAddr = detectProxy();
+        ResolutionResult.Builder resolutionResultBuilder = ResolutionResult.newBuilder();
+        if (proxiedAddr != null) {
+          if (logger.isLoggable(Level.FINER)) {
+            logger.finer("Using proxy address " + proxiedAddr);
+          }
+          resolutionResultBuilder.setAddresses(Collections.singletonList(proxiedAddr));
+        } else {
+          result = doResolve(false);
+          if (result.error != null) {
+            savedListener.onError(result.error);
+            return;
+          }
+          if (result.addresses != null) {
+            resolutionResultBuilder.setAddresses(result.addresses);
+          }
+          if (result.config != null) {
+            resolutionResultBuilder.setServiceConfig(result.config);
+          }
+          if (result.attributes != null) {
+            resolutionResultBuilder.setAttributes(result.attributes);
+          }
+        }
+        savedListener.onResult(resolutionResultBuilder.build());
       } catch (IOException e) {
         savedListener.onError(
             Status.UNAVAILABLE.withDescription("Unable to resolve host " + host).withCause(e));
-        return;
-      }
-      if (proxiedAddr != null) {
-        if (logger.isLoggable(Level.FINER)) {
-          logger.finer("Using proxy address " + proxiedAddr);
-        }
-        EquivalentAddressGroup server = new EquivalentAddressGroup(proxiedAddr);
-        ResolutionResult resolutionResult =
-            ResolutionResult.newBuilder()
-                .setAddresses(Collections.singletonList(server))
-                .setAttributes(Attributes.EMPTY)
-                .build();
-        savedListener.onResult(resolutionResult);
-        return;
-      }
-
-      ResolutionResults resolutionResults;
-      try {
-        ResourceResolver resourceResolver = null;
-        if (shouldUseJndi(enableJndi, enableJndiLocalhost, host)) {
-          resourceResolver = getResourceResolver();
-        }
-        final ResolutionResults results = resolveAll(
-            addressResolver,
-            resourceResolver,
-            enableSrv,
-            enableTxt,
-            host);
-        resolutionResults = results;
+      } finally {
+        final boolean succeed = result != null && result.error == null;
         syncContext.execute(new Runnable() {
-            @Override
-            public void run() {
-              cachedResolutionResults = results;
+          @Override
+          public void run() {
+            if (succeed) {
+              resolved = true;
               if (cacheTtlNanos > 0) {
                 stopwatch.reset().start();
               }
             }
-          });
-        if (logger.isLoggable(Level.FINER)) {
-          logger.finer("Found DNS results " + resolutionResults + " for " + host);
-        }
-      } catch (Exception e) {
-        savedListener.onError(
-            Status.UNAVAILABLE.withDescription("Unable to resolve host " + host).withCause(e));
-        return;
-      }
-      // Each address forms an EAG
-      List<EquivalentAddressGroup> servers = new ArrayList<>();
-      for (InetAddress inetAddr : resolutionResults.addresses) {
-        servers.add(new EquivalentAddressGroup(new InetSocketAddress(inetAddr, port)));
-      }
-      servers.addAll(resolutionResults.balancerAddresses);
-      if (servers.isEmpty()) {
-        savedListener.onError(Status.UNAVAILABLE.withDescription(
-            "No DNS backend or balancer addresses found for " + host));
-        return;
-      }
-
-      Attributes.Builder attrs = Attributes.newBuilder();
-      if (!resolutionResults.txtRecords.isEmpty()) {
-        ConfigOrError serviceConfig =
-            parseServiceConfig(resolutionResults.txtRecords, random, getLocalHostname());
-        if (serviceConfig != null) {
-          if (serviceConfig.getError() != null) {
-            savedListener.onError(serviceConfig.getError());
-            return;
-          } else {
-            @SuppressWarnings("unchecked")
-            Map<String, ?> config = (Map<String, ?>) serviceConfig.getConfig();
-            attrs.set(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG, config);
+            resolving = false;
           }
-        }
-      } else {
-        logger.log(Level.FINE, "No TXT records found for {0}", new Object[]{host});
+        });
       }
-      ResolutionResult resolutionResult =
-          ResolutionResult.newBuilder().setAddresses(servers).setAttributes(attrs.build()).build();
-      savedListener.onResult(resolutionResult);
     }
   }
 
@@ -364,7 +390,7 @@ final class DnsNameResolver extends NameResolver {
   }
 
   private boolean cacheRefreshRequired() {
-    return cachedResolutionResults == null
+    return !resolved
         || cacheTtlNanos == 0
         || (cacheTtlNanos > 0 && stopwatch.elapsed(TimeUnit.NANOSECONDS) > cacheTtlNanos);
   }
@@ -384,69 +410,6 @@ final class DnsNameResolver extends NameResolver {
     return port;
   }
 
-  @VisibleForTesting
-  static ResolutionResults resolveAll(
-      AddressResolver addressResolver,
-      @Nullable ResourceResolver resourceResolver,
-      boolean requestSrvRecords,
-      boolean requestTxtRecords,
-      String name) {
-    List<? extends InetAddress> addresses = Collections.emptyList();
-    Exception addressesException = null;
-    List<EquivalentAddressGroup> balancerAddresses = Collections.emptyList();
-    Exception balancerAddressesException = null;
-    List<String> txtRecords = Collections.emptyList();
-    Exception txtRecordsException = null;
-
-    try {
-      addresses = addressResolver.resolveAddress(name);
-    } catch (Exception e) {
-      addressesException = e;
-    }
-    if (resourceResolver != null) {
-      if (requestSrvRecords) {
-        try {
-          balancerAddresses =
-              resourceResolver.resolveSrv(addressResolver, GRPCLB_NAME_PREFIX + name);
-        } catch (Exception e) {
-          balancerAddressesException = e;
-        }
-      }
-      if (requestTxtRecords) {
-        boolean balancerLookupFailedOrNotAttempted =
-            !requestSrvRecords || balancerAddressesException != null;
-        boolean dontResolveTxt =
-            (addressesException != null) && balancerLookupFailedOrNotAttempted;
-        // Only do the TXT record lookup if one of the above address resolutions succeeded.
-        if (!dontResolveTxt) {
-          try {
-            txtRecords = resourceResolver.resolveTxt(SERVICE_CONFIG_NAME_PREFIX + name);
-          } catch (Exception e) {
-            txtRecordsException = e;
-          }
-        }
-      }
-    }
-    try {
-      if (addressesException != null
-          && (balancerAddressesException != null || balancerAddresses.isEmpty())) {
-        Throwables.throwIfUnchecked(addressesException);
-        throw new RuntimeException(addressesException);
-      }
-    } finally {
-      if (addressesException != null) {
-        logger.log(Level.FINE, "Address resolution failure", addressesException);
-      }
-      if (balancerAddressesException != null) {
-        logger.log(Level.FINE, "Balancer resolution failure", balancerAddressesException);
-      }
-      if (txtRecordsException != null) {
-        logger.log(Level.FINE, "ServiceConfig resolution failure", txtRecordsException);
-      }
-    }
-    return new ResolutionResults(addresses, txtRecords, balancerAddresses);
-  }
-
   /**
    *
    * @throws IOException if one of the txt records contains improperly formatted JSON.
@@ -471,29 +434,20 @@ final class DnsNameResolver extends NameResolver {
 
   @Nullable
   private static final Double getPercentageFromChoice(Map<String, ?> serviceConfigChoice) {
-    if (!serviceConfigChoice.containsKey(SERVICE_CONFIG_CHOICE_PERCENTAGE_KEY)) {
-      return null;
-    }
-    return JsonUtil.getDouble(serviceConfigChoice, SERVICE_CONFIG_CHOICE_PERCENTAGE_KEY);
+    return JsonUtil.getNumber(serviceConfigChoice, SERVICE_CONFIG_CHOICE_PERCENTAGE_KEY);
   }
 
   @Nullable
   private static final List<String> getClientLanguagesFromChoice(
       Map<String, ?> serviceConfigChoice) {
-    if (!serviceConfigChoice.containsKey(SERVICE_CONFIG_CHOICE_CLIENT_LANGUAGE_KEY)) {
-      return null;
-    }
-    return JsonUtil.checkStringList(
-        JsonUtil.getList(serviceConfigChoice, SERVICE_CONFIG_CHOICE_CLIENT_LANGUAGE_KEY));
+    return JsonUtil.getListOfStrings(
+        serviceConfigChoice, SERVICE_CONFIG_CHOICE_CLIENT_LANGUAGE_KEY);
   }
 
   @Nullable
   private static final List<String> getHostnamesFromChoice(Map<String, ?> serviceConfigChoice) {
-    if (!serviceConfigChoice.containsKey(SERVICE_CONFIG_CHOICE_CLIENT_HOSTNAME_KEY)) {
-      return null;
-    }
-    return JsonUtil.checkStringList(
-        JsonUtil.getList(serviceConfigChoice, SERVICE_CONFIG_CHOICE_CLIENT_HOSTNAME_KEY));
+    return JsonUtil.getListOfStrings(
+        serviceConfigChoice, SERVICE_CONFIG_CHOICE_CLIENT_HOSTNAME_KEY);
   }
 
   /**
@@ -533,7 +487,7 @@ final class DnsNameResolver extends NameResolver {
   @VisibleForTesting
   static Map<String, ?> maybeChooseServiceConfig(
       Map<String, ?> choice, Random random, String hostname) {
-    for (Entry<String, ?> entry : choice.entrySet()) {
+    for (Map.Entry<String, ?> entry : choice.entrySet()) {
       Verify.verify(SERVICE_CONFIG_CHOICE_KEYS.contains(entry.getKey()), "Bad key: %s", entry);
     }
 
@@ -581,41 +535,64 @@ final class DnsNameResolver extends NameResolver {
   }
 
   /**
-   * Describes the results from a DNS query.
+   * Used as a DNS-based name resolver's internal representation of resolution result.
+   */
+  protected static final class InternalResolutionResult {
+    private Status error;
+    private List<EquivalentAddressGroup> addresses;
+    private ConfigOrError config;
+    public Attributes attributes;
+
+    private InternalResolutionResult() {}
+  }
+
+  /**
+   * Describes a parsed SRV record.
    */
   @VisibleForTesting
-  static final class ResolutionResults {
-    final List<? extends InetAddress> addresses;
-    final List<String> txtRecords;
-    final List<EquivalentAddressGroup> balancerAddresses;
+  public static final class SrvRecord {
+    public final String host;
+    public final int port;
 
-    ResolutionResults(
-        List<? extends InetAddress> addresses,
-        List<String> txtRecords,
-        List<EquivalentAddressGroup> balancerAddresses) {
-      this.addresses = Collections.unmodifiableList(checkNotNull(addresses, "addresses"));
-      this.txtRecords = Collections.unmodifiableList(checkNotNull(txtRecords, "txtRecords"));
-      this.balancerAddresses =
-          Collections.unmodifiableList(checkNotNull(balancerAddresses, "balancerAddresses"));
+    public SrvRecord(String host, int port) {
+      this.host = host;
+      this.port = port;
+    }
+
+    @Override
+    public int hashCode() {
+      return Objects.hashCode(host, port);
+    }
+
+    @Override
+    public boolean equals(Object obj) {
+      if (this == obj) {
+        return true;
+      }
+      if (obj == null || getClass() != obj.getClass()) {
+        return false;
+      }
+      SrvRecord that = (SrvRecord) obj;
+      return port == that.port && host.equals(that.host);
     }
 
     @Override
     public String toString() {
-      return MoreObjects.toStringHelper(this)
-          .add("addresses", addresses)
-          .add("txtRecords", txtRecords)
-          .add("balancerAddresses", balancerAddresses)
-          .toString();
+      return
+          MoreObjects.toStringHelper(this)
+              .add("host", host)
+              .add("port", port)
+              .toString();
     }
   }
 
   @VisibleForTesting
-  void setAddressResolver(AddressResolver addressResolver) {
+  protected void setAddressResolver(AddressResolver addressResolver) {
     this.addressResolver = addressResolver;
   }
 
   @VisibleForTesting
-  void setResourceResolver(ResourceResolver resourceResolver) {
+  protected void setResourceResolver(ResourceResolver resourceResolver) {
     this.resourceResolver.set(resourceResolver);
   }
 
@@ -641,7 +618,8 @@ final class DnsNameResolver extends NameResolver {
   /**
    * AddressResolver resolves a hostname into a list of addresses.
    */
-  interface AddressResolver {
+  @VisibleForTesting
+  public interface AddressResolver {
     List<InetAddress> resolveAddress(String host) throws Exception;
   }
 
@@ -657,15 +635,18 @@ final class DnsNameResolver extends NameResolver {
   /**
    * {@link ResourceResolver} is a Dns ResourceRecord resolver.
    */
-  interface ResourceResolver {
+  @VisibleForTesting
+  public interface ResourceResolver {
     List<String> resolveTxt(String host) throws Exception;
 
-    List<EquivalentAddressGroup> resolveSrv(
-        AddressResolver addressResolver, String host) throws Exception;
+    List<SrvRecord> resolveSrv(String host) throws Exception;
   }
 
   @Nullable
-  private ResourceResolver getResourceResolver() {
+  protected ResourceResolver getResourceResolver() {
+    if (!shouldUseJndi(enableJndi, enableJndiLocalhost, host)) {
+      return null;
+    }
     ResourceResolver rr;
     if ((rr = resourceResolver.get()) == null) {
       if (resourceResolverFactory != null) {
@@ -733,7 +714,8 @@ final class DnsNameResolver extends NameResolver {
   }
 
   @VisibleForTesting
-  static boolean shouldUseJndi(boolean jndiEnabled, boolean jndiLocalhostEnabled, String target) {
+  protected static boolean shouldUseJndi(
+      boolean jndiEnabled, boolean jndiLocalhostEnabled, String target) {
     if (!jndiEnabled) {
       return false;
     }
