@@ -29,11 +29,6 @@ abstract class ForwardingClientStreamListener implements ClientStreamListener {
     delegate().headersRead(headers);
   }
 
-  @Override
-  public void closed(Status status, Metadata trailers) {
-    delegate().closed(status, trailers);
-  }
-
   @Override
   public void closed(Status status, RpcProgress rpcProgress, Metadata trailers) {
     delegate().closed(status, rpcProgress, trailers);
