@@ -34,6 +34,7 @@ import io.grpc.Attributes;
 import io.grpc.CallOptions;
 import io.grpc.ChannelLogger;
 import io.grpc.ChannelLogger.ChannelLogLevel;
+import io.grpc.ClientStreamTracer;
 import io.grpc.ConnectivityState;
 import io.grpc.ConnectivityStateInfo;
 import io.grpc.EquivalentAddressGroup;
@@ -111,6 +112,10 @@ final class InternalSubchannel implements InternalInstrumented<ChannelStats>, Tr
 
   @Nullable
   private ScheduledHandle reconnectTask;
+  @Nullable
+  private ScheduledHandle shutdownDueToUpdateTask;
+  @Nullable
+  private ManagedClientTransport shutdownDueToUpdateTransport;
 
   /**
    * All transports that are not terminated. At the very least the value of {@link #activeTransport}
@@ -335,12 +340,12 @@ final class InternalSubchannel implements InternalInstrumented<ChannelStats>, Tr
     Preconditions.checkNotNull(newAddressGroups, "newAddressGroups");
     checkListHasNoNulls(newAddressGroups, "newAddressGroups contains null entry");
     Preconditions.checkArgument(!newAddressGroups.isEmpty(), "newAddressGroups is empty");
+    final List<EquivalentAddressGroup> newImmutableAddressGroups =
+        Collections.unmodifiableList(new ArrayList<>(newAddressGroups));
 
     syncContext.execute(new Runnable() {
       @Override
       public void run() {
-        List<EquivalentAddressGroup> newImmutableAddressGroups =
-            Collections.unmodifiableList(new ArrayList<>(newAddressGroups));
         ManagedClientTransport savedTransport = null;
         SocketAddress previousAddress = addressIndex.getCurrentAddress();
         addressIndex.updateGroups(newImmutableAddressGroups);
@@ -354,7 +359,9 @@ final class InternalSubchannel implements InternalInstrumented<ChannelStats>, Tr
               addressIndex.reset();
               gotoNonErrorState(IDLE);
             } else {
-              savedTransport = pendingTransport;
+              pendingTransport.shutdown(
+                  Status.UNAVAILABLE.withDescription(
+                    "InternalSubchannel closed pending transport due to address change"));
               pendingTransport = null;
               addressIndex.reset();
               startNewTransport();
@@ -362,9 +369,33 @@ final class InternalSubchannel implements InternalInstrumented<ChannelStats>, Tr
           }
         }
         if (savedTransport != null) {
-          savedTransport.shutdown(
-              Status.UNAVAILABLE.withDescription(
-                  "InternalSubchannel closed transport due to address change"));
+          if (shutdownDueToUpdateTask != null) {
+            // Keeping track of multiple shutdown tasks adds complexity, and shouldn't generally be
+            // necessary. This transport has probably already had plenty of time.
+            shutdownDueToUpdateTransport.shutdown(
+                Status.UNAVAILABLE.withDescription(
+                    "InternalSubchannel closed transport early due to address change"));
+            shutdownDueToUpdateTask.cancel();
+            shutdownDueToUpdateTask = null;
+            shutdownDueToUpdateTransport = null;
+          }
+          // Avoid needless RPC failures by delaying the shutdown. See
+          // https://github.com/grpc/grpc-java/issues/2562
+          shutdownDueToUpdateTransport = savedTransport;
+          shutdownDueToUpdateTask = syncContext.schedule(
+              new Runnable() {
+                @Override public void run() {
+                  ManagedClientTransport transport = shutdownDueToUpdateTransport;
+                  shutdownDueToUpdateTask = null;
+                  shutdownDueToUpdateTransport = null;
+                  transport.shutdown(
+                      Status.UNAVAILABLE.withDescription(
+                          "InternalSubchannel closed transport due to address change"));
+                }
+              },
+              ManagedChannelImpl.SUBCHANNEL_SHUTDOWN_DELAY_SECONDS,
+              TimeUnit.SECONDS,
+              scheduledExecutor);
         }
       }
     });
@@ -390,6 +421,12 @@ final class InternalSubchannel implements InternalInstrumented<ChannelStats>, Tr
           handleTermination();
         }  // else: the callback will be run once all transports have been terminated
         cancelReconnectTask();
+        if (shutdownDueToUpdateTask != null) {
+          shutdownDueToUpdateTask.cancel();
+          shutdownDueToUpdateTransport.shutdown(reason);
+          shutdownDueToUpdateTask = null;
+          shutdownDueToUpdateTransport = null;
+        }
         if (savedActiveTransport != null) {
           savedActiveTransport.shutdown(reason);
         }
@@ -631,8 +668,9 @@ final class InternalSubchannel implements InternalInstrumented<ChannelStats>, Tr
 
     @Override
     public ClientStream newStream(
-        MethodDescriptor<?, ?> method, Metadata headers, CallOptions callOptions) {
-      final ClientStream streamDelegate = super.newStream(method, headers, callOptions);
+        MethodDescriptor<?, ?> method, Metadata headers, CallOptions callOptions,
+        ClientStreamTracer[] tracers) {
+      final ClientStream streamDelegate = super.newStream(method, headers, callOptions, tracers);
       return new ForwardingClientStream() {
         @Override
         protected ClientStream delegate() {
@@ -648,12 +686,6 @@ final class InternalSubchannel implements InternalInstrumented<ChannelStats>, Tr
               return listener;
             }
 
-            @Override
-            public void closed(Status status, Metadata trailers) {
-              callTracer.reportCallEnded(status.isOk());
-              super.closed(status, trailers);
-            }
-
             @Override
             public void closed(
                 Status status, RpcProgress rpcProgress, Metadata trailers) {
