@@ -75,6 +75,7 @@ import org.mockito.stubbing.Answer;
 public class AbstractClientStreamTest {
 
   @Rule public final MockitoRule mocks = MockitoJUnit.rule();
+  @SuppressWarnings("deprecation") // https://github.com/grpc/grpc-java/issues/7467
   @Rule public final ExpectedException thrown = ExpectedException.none();
 
   private final StatsTraceContext statsTraceCtx = StatsTraceContext.NOOP;
@@ -333,9 +334,8 @@ public class AbstractClientStreamTest {
     AbstractClientStream stream =
         new BaseAbstractClientStream(allocator, statsTraceCtx, transportTracer);
     stream.start(mockListener);
-    // The application will call request when waiting for a message, which will in turn call this
-    // on the transport thread.
-    stream.transportState().requestMessagesFromDeframer(1);
+    // The application will call request when waiting for a message
+    stream.request(1);
     // Send first byte of 2 byte message
     stream.transportState().deframe(ReadableBuffers.wrap(new byte[] {0, 0, 0, 0, 2, 1}));
     Status status = Status.INTERNAL.withDescription("rst___stream");
@@ -359,7 +359,7 @@ public class AbstractClientStreamTest {
     Status status = Status.INTERNAL.withDescription("rst___stream");
     // Simulate getting a reset
     stream.transportState().transportReportStatus(status, false /*stop delivery*/, new Metadata());
-    stream.transportState().requestMessagesFromDeframer(1);
+    stream.request(1);
 
     ArgumentCaptor<Status> statusCaptor = ArgumentCaptor.forClass(Status.class);
     verify(mockListener)
@@ -373,7 +373,7 @@ public class AbstractClientStreamTest {
         new BaseAbstractClientStream(allocator, statsTraceCtx, transportTracer);
     stream.start(mockListener);
 
-    stream.transportState().requestMessagesFromDeframer(1);
+    stream.request(1);
     stream.transportState().deframe(ReadableBuffers.wrap(new byte[] {0, 0, 0, 0, 2, 1}));
     stream.transportState().inboundTrailersReceived(new Metadata(), Status.OK);
 
@@ -390,7 +390,7 @@ public class AbstractClientStreamTest {
         new BaseAbstractClientStream(allocator, statsTraceCtx, transportTracer);
     stream.start(mockListener);
 
-    stream.transportState().requestMessagesFromDeframer(1);
+    stream.request(1);
     stream.transportState().deframe(ReadableBuffers.wrap(new byte[] {0, 0, 0, 0, 2, 1}));
     stream.transportState().inboundTrailersReceived(
         new Metadata(), Status.DATA_LOSS.withDescription("data___loss"));
@@ -551,9 +551,6 @@ public class AbstractClientStreamTest {
     @Override
     public void writeHeaders(Metadata headers, byte[] payload) {}
 
-    @Override
-    public void request(int numMessages) {}
-
     @Override
     public void writeFrame(
         WritableBuffer frame, boolean endOfStream, boolean flush, int numMessages) {}
