@@ -19,6 +19,8 @@ package io.grpc.internal;
 import static com.google.common.truth.Truth.assertThat;
 import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertFalse;
+import static org.junit.Assert.assertNotNull;
+import static org.junit.Assert.assertNull;
 import static org.junit.Assert.assertTrue;
 import static org.mockito.ArgumentMatchers.any;
 import static org.mockito.ArgumentMatchers.eq;
@@ -33,14 +35,17 @@ import static org.mockito.Mockito.verifyNoMoreInteractions;
 import static org.mockito.Mockito.when;
 
 import io.grpc.Attributes;
-import io.grpc.Attributes.Key;
 import io.grpc.Codec;
+import io.grpc.Compressor;
+import io.grpc.Deadline;
 import io.grpc.DecompressorRegistry;
 import io.grpc.Metadata;
 import io.grpc.Status;
+import io.grpc.internal.ClientStreamListener.RpcProgress;
 import io.grpc.internal.testing.SingleMessageProducer;
 import java.io.ByteArrayInputStream;
 import java.io.InputStream;
+import java.util.concurrent.TimeUnit;
 import org.junit.Rule;
 import org.junit.Test;
 import org.junit.runner.RunWith;
@@ -73,7 +78,7 @@ public class DelayedStreamTest {
     final String authority = "becauseIsaidSo";
     stream.setAuthority(authority);
     stream.start(listener);
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
     InOrder inOrder = inOrder(realStream);
     inOrder.verify(realStream).setAuthority(authority);
     inOrder.verify(realStream).start(any(ClientStreamListener.class));
@@ -91,11 +96,32 @@ public class DelayedStreamTest {
     stream.start(mock(ClientStreamListener.class));
   }
 
+  @Test(expected = IllegalStateException.class)
+  public void writeMessage_beforeStart() {
+    InputStream message = new ByteArrayInputStream(new byte[]{'a'});
+    stream.writeMessage(message);
+  }
+
+  @Test(expected = IllegalStateException.class)
+  public void flush_beforeStart() {
+    stream.flush();
+  }
+
+  @Test(expected = IllegalStateException.class)
+  public void request_beforeStart() {
+    stream.request(1);
+  }
+
+  @Test(expected = IllegalStateException.class)
+  public void halfClose_beforeStart() {
+    stream.halfClose();
+  }
+
   @Test
   public void setStream_sendsAllMessages() {
-    stream.start(listener);
     stream.setCompressor(Codec.Identity.NONE);
     stream.setDecompressorRegistry(DecompressorRegistry.getDefaultInstance());
+    stream.start(listener);
 
     stream.setMessageCompression(true);
     InputStream message = new ByteArrayInputStream(new byte[]{'a'});
@@ -103,7 +129,7 @@ public class DelayedStreamTest {
     stream.setMessageCompression(false);
     stream.writeMessage(message);
 
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
 
     verify(realStream).setCompressor(Codec.Identity.NONE);
     verify(realStream).setDecompressorRegistry(DecompressorRegistry.getDefaultInstance());
@@ -126,7 +152,7 @@ public class DelayedStreamTest {
   public void setStream_halfClose() {
     stream.start(listener);
     stream.halfClose();
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
 
     verify(realStream).halfClose();
   }
@@ -135,7 +161,7 @@ public class DelayedStreamTest {
   public void setStream_flush() {
     stream.start(listener);
     stream.flush();
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
     verify(realStream).flush();
 
     stream.flush();
@@ -147,7 +173,7 @@ public class DelayedStreamTest {
     stream.start(listener);
     stream.request(1);
     stream.request(2);
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
     verify(realStream).request(1);
     verify(realStream).request(2);
 
@@ -155,11 +181,99 @@ public class DelayedStreamTest {
     verify(realStream).request(3);
   }
 
+  @Test
+  public void setStreamThenStart() {
+    stream.optimizeForDirectExecutor();
+    stream.setCompressor(mock(Compressor.class));
+    stream.setFullStreamDecompression(false);
+    stream.setDecompressorRegistry(DecompressorRegistry.emptyInstance());
+    stream.setDeadline(Deadline.after(1, TimeUnit.MINUTES));
+    stream.setAuthority("auth");
+    stream.setMaxInboundMessageSize(10);
+    stream.setMaxOutboundMessageSize(10);
+
+    assertNull(stream.setStream(realStream));
+    stream.start(listener);
+    stream.request(1);
+
+    InOrder inOrder = inOrder(realStream);
+    inOrder.verify(realStream).optimizeForDirectExecutor();
+    inOrder.verify(realStream).setCompressor(any(Compressor.class));
+    inOrder.verify(realStream).setFullStreamDecompression(false);
+    inOrder.verify(realStream).setDecompressorRegistry(any(DecompressorRegistry.class));
+    inOrder.verify(realStream).setDeadline(any(Deadline.class));
+    inOrder.verify(realStream).setAuthority("auth");
+    inOrder.verify(realStream).setMaxInboundMessageSize(10);
+    inOrder.verify(realStream).setMaxOutboundMessageSize(10);
+    verify(realStream).request(1);
+    verify(realStream).start(same(listener));
+    verifyNoMoreInteractions(realStream);
+  }
+
+  @Test
+  public void startThenSetRealStream() {
+    stream.setAuthority("auth");
+    stream.optimizeForDirectExecutor();
+    stream.setMaxInboundMessageSize(10);
+    stream.setCompressor(mock(Compressor.class));
+    stream.setFullStreamDecompression(false);
+    stream.setDeadline(Deadline.after(1, TimeUnit.MINUTES));
+    stream.setMaxOutboundMessageSize(10);
+    stream.setDecompressorRegistry(DecompressorRegistry.emptyInstance());
+    stream.start(listener);
+    stream.request(1);
+    InputStream message = mock(InputStream.class);
+    stream.writeMessage(message);
+    Runnable runnable = stream.setStream(realStream);
+    assertNotNull(runnable);
+    callMeMaybe(runnable);
+    stream.getAttributes();
+    stream.request(4);
+
+    InOrder inOrder = inOrder(realStream);
+    inOrder.verify(realStream).setAuthority("auth");
+    inOrder.verify(realStream).optimizeForDirectExecutor();
+    inOrder.verify(realStream).setMaxInboundMessageSize(10);
+    inOrder.verify(realStream).setCompressor(any(Compressor.class));
+    inOrder.verify(realStream).setFullStreamDecompression(false);
+    inOrder.verify(realStream).setDeadline(any(Deadline.class));
+    inOrder.verify(realStream).setMaxOutboundMessageSize(10);
+    inOrder.verify(realStream).setDecompressorRegistry(any(DecompressorRegistry.class));
+    inOrder.verify(realStream).start(listenerCaptor.capture());
+    inOrder.verify(realStream).request(1);
+    inOrder.verify(realStream).writeMessage(same(message));
+    inOrder.verify(realStream).getAttributes();
+    verify(realStream).request(4);
+    verifyNoMoreInteractions(realStream);
+    ClientStreamListener delayedListener = listenerCaptor.getValue();
+    delayedListener.onReady();
+    verify(listener).onReady();
+  }
+
+  @Test
+  public void drainPendingCallRacesCancel() {
+    stream.start(listener);
+    InputStream message = mock(InputStream.class);
+    stream.writeMessage(message);
+    stream.flush();
+    Runnable runnable = stream.setStream(realStream);
+    assertNotNull(runnable);
+    stream.cancel(Status.CANCELLED);
+    callMeMaybe(runnable);
+
+    InOrder inOrder = inOrder(realStream);
+    inOrder.verify(realStream).start(any(ClientStreamListener.class));
+    inOrder.verify(realStream).writeMessage(same(message));
+    inOrder.verify(realStream).flush();
+    inOrder.verify(realStream).cancel(Status.CANCELLED);
+    verifyNoMoreInteractions(realStream);
+  }
+
   @Test
   public void setStream_setMessageCompression() {
     stream.start(listener);
     stream.setMessageCompression(false);
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
     verify(realStream).setMessageCompression(false);
 
     stream.setMessageCompression(true);
@@ -170,7 +284,7 @@ public class DelayedStreamTest {
   public void setStream_isReady() {
     stream.start(listener);
     assertFalse(stream.isReady());
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
     verify(realStream, never()).isReady();
 
     assertFalse(stream.isReady());
@@ -184,14 +298,14 @@ public class DelayedStreamTest {
   @Test
   public void setStream_getAttributes() {
     Attributes attributes =
-        Attributes.newBuilder().set(Key.<String>create("fakeKey"), "fakeValue").build();
+        Attributes.newBuilder().set(Attributes.Key.<String>create("fakeKey"), "fakeValue").build();
     when(realStream.getAttributes()).thenReturn(attributes);
 
     stream.start(listener);
 
     assertEquals(Attributes.EMPTY, stream.getAttributes());
 
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
     assertEquals(attributes, stream.getAttributes());
   }
 
@@ -199,13 +313,13 @@ public class DelayedStreamTest {
   public void startThenCancelled() {
     stream.start(listener);
     stream.cancel(Status.CANCELLED);
-    verify(listener).closed(eq(Status.CANCELLED), any(Metadata.class));
+    verify(listener).closed(eq(Status.CANCELLED), any(RpcProgress.class), any(Metadata.class));
   }
 
   @Test
   public void startThenSetStreamThenCancelled() {
     stream.start(listener);
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
     stream.cancel(Status.CANCELLED);
     verify(realStream).start(any(ClientStreamListener.class));
     verify(realStream).cancel(same(Status.CANCELLED));
@@ -213,52 +327,36 @@ public class DelayedStreamTest {
 
   @Test
   public void setStreamThenStartThenCancelled() {
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
     stream.start(listener);
     stream.cancel(Status.CANCELLED);
     verify(realStream).start(same(listener));
     verify(realStream).cancel(same(Status.CANCELLED));
   }
 
-  @Test
-  public void setStreamThenCancelled() {
-    stream.setStream(realStream);
-    stream.cancel(Status.CANCELLED);
-    verify(realStream).cancel(same(Status.CANCELLED));
-  }
-
   @Test
   public void setStreamTwice() {
     stream.start(listener);
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
     verify(realStream).start(any(ClientStreamListener.class));
-    stream.setStream(mock(ClientStream.class));
+    callMeMaybe(stream.setStream(mock(ClientStream.class)));
     stream.flush();
     verify(realStream).flush();
   }
 
   @Test
   public void cancelThenSetStream() {
-    stream.cancel(Status.CANCELLED);
-    stream.setStream(realStream);
     stream.start(listener);
+    stream.cancel(Status.CANCELLED);
+    callMeMaybe(stream.setStream(realStream));
     stream.isReady();
     verifyNoMoreInteractions(realStream);
   }
 
-  @Test
+  @Test(expected = IllegalStateException.class)
   public void cancel_beforeStart() {
     Status status = Status.CANCELLED.withDescription("that was quick");
     stream.cancel(status);
-    stream.start(listener);
-    verify(listener).closed(same(status), any(Metadata.class));
-  }
-
-  @Test
-  public void cancelledThenStart() {
-    stream.cancel(Status.CANCELLED);
-    stream.start(listener);
-    verify(listener).closed(eq(Status.CANCELLED), any(Metadata.class));
   }
 
   @Test
@@ -276,7 +374,7 @@ public class DelayedStreamTest {
 
     IsReadyListener isReadyListener = new IsReadyListener();
     stream.start(isReadyListener);
-    stream.setStream(new NoopClientStream() {
+    callMeMaybe(stream.setStream(new NoopClientStream() {
       @Override
       public void start(ClientStreamListener listener) {
         // This call to the listener should end up being delayed.
@@ -287,7 +385,7 @@ public class DelayedStreamTest {
       public boolean isReady() {
         return true;
       }
-    });
+    }));
     assertTrue(isReadyListener.onReadyCalled);
   }
 
@@ -303,7 +401,7 @@ public class DelayedStreamTest {
 
     final InOrder inOrder = inOrder(listener);
     stream.start(listener);
-    stream.setStream(new NoopClientStream() {
+    callMeMaybe(stream.setStream(new NoopClientStream() {
       @Override
       public void start(ClientStreamListener passedListener) {
         passedListener.onReady();
@@ -311,17 +409,17 @@ public class DelayedStreamTest {
         passedListener.messagesAvailable(producer1);
         passedListener.onReady();
         passedListener.messagesAvailable(producer2);
-        passedListener.closed(status, trailers);
+        passedListener.closed(status, RpcProgress.PROCESSED, trailers);
 
         verifyNoMoreInteractions(listener);
       }
-    });
+    }));
     inOrder.verify(listener).onReady();
     inOrder.verify(listener).headersRead(headers);
     inOrder.verify(listener).messagesAvailable(producer1);
     inOrder.verify(listener).onReady();
     inOrder.verify(listener).messagesAvailable(producer2);
-    inOrder.verify(listener).closed(status, trailers);
+    inOrder.verify(listener).closed(status, RpcProgress.PROCESSED, trailers);
   }
 
   @Test
@@ -333,7 +431,7 @@ public class DelayedStreamTest {
     final Status status = Status.UNKNOWN.withDescription("unique status");
 
     stream.start(listener);
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
     verify(realStream).start(listenerCaptor.capture());
     ClientStreamListener delayedListener = listenerCaptor.getValue();
     delayedListener.onReady();
@@ -342,8 +440,8 @@ public class DelayedStreamTest {
     verify(listener).headersRead(headers);
     delayedListener.messagesAvailable(producer);
     verify(listener).messagesAvailable(producer);
-    delayedListener.closed(status, trailers);
-    verify(listener).closed(status, trailers);
+    delayedListener.closed(status, RpcProgress.PROCESSED, trailers);
+    verify(listener).closed(status, RpcProgress.PROCESSED, trailers);
   }
 
   @Test
@@ -372,11 +470,17 @@ public class DelayedStreamTest {
         }
       }).when(realStream).appendTimeoutInsight(any(InsightBuilder.class));
     stream.start(listener);
-    stream.setStream(realStream);
+    callMeMaybe(stream.setStream(realStream));
 
     InsightBuilder insight = new InsightBuilder();
     stream.appendTimeoutInsight(insight);
     assertThat(insight.toString())
         .matches("\\[buffered_nanos=[0-9]+, remote_addr=127\\.0\\.0\\.1:443\\]");
   }
+
+  private void callMeMaybe(Runnable r) {
+    if (r != null) {
+      r.run();
+    }
+  }
 }
