@@ -337,6 +337,7 @@ public class MessageDeframerTest {
 
   @RunWith(JUnit4.class)
   public static class SizeEnforcingInputStreamTests {
+    @SuppressWarnings("deprecation") // https://github.com/grpc/grpc-java/issues/7467
     @Rule
     public final ExpectedException thrown = ExpectedException.none();
 
@@ -377,7 +378,7 @@ public class MessageDeframerTest {
 
       try {
         thrown.expect(StatusRuntimeException.class);
-        thrown.expectMessage("RESOURCE_EXHAUSTED: Compressed gRPC message exceeds");
+        thrown.expectMessage("RESOURCE_EXHAUSTED: Decompressed gRPC message exceeds");
 
         while (stream.read() != -1) {
         }
@@ -423,7 +424,7 @@ public class MessageDeframerTest {
 
       try {
         thrown.expect(StatusRuntimeException.class);
-        thrown.expectMessage("RESOURCE_EXHAUSTED: Compressed gRPC message exceeds");
+        thrown.expectMessage("RESOURCE_EXHAUSTED: Decompressed gRPC message exceeds");
 
         stream.read(buf, 0, buf.length);
       } finally {
@@ -466,7 +467,7 @@ public class MessageDeframerTest {
 
       try {
         thrown.expect(StatusRuntimeException.class);
-        thrown.expectMessage("RESOURCE_EXHAUSTED: Compressed gRPC message exceeds");
+        thrown.expectMessage("RESOURCE_EXHAUSTED: Decompressed gRPC message exceeds");
 
         stream.skip(4);
       } finally {
