@@ -44,7 +44,6 @@ import static org.mockito.Mockito.verify;
 import static org.mockito.Mockito.verifyNoMoreInteractions;
 import static org.mockito.Mockito.when;
 
-import com.google.common.collect.ImmutableList;
 import com.google.common.util.concurrent.ListenableFuture;
 import com.google.common.util.concurrent.MoreExecutors;
 import com.google.common.util.concurrent.SettableFuture;
@@ -66,6 +65,7 @@ import io.grpc.Metadata;
 import io.grpc.MethodDescriptor;
 import io.grpc.ServerCall;
 import io.grpc.ServerCall.Listener;
+import io.grpc.ServerCallExecutorSupplier;
 import io.grpc.ServerCallHandler;
 import io.grpc.ServerInterceptor;
 import io.grpc.ServerMethodDefinition;
@@ -74,14 +74,15 @@ import io.grpc.ServerStreamTracer;
 import io.grpc.ServerTransportFilter;
 import io.grpc.ServiceDescriptor;
 import io.grpc.Status;
+import io.grpc.Status.Code;
 import io.grpc.StringMarshaller;
 import io.grpc.internal.ServerImpl.JumpToApplicationThreadServerStreamListener;
+import io.grpc.internal.ServerImplBuilder.ClientTransportServersBuilder;
 import io.grpc.internal.testing.SingleMessageProducer;
 import io.grpc.internal.testing.TestServerStreamTracer;
 import io.grpc.util.MutableHandlerRegistry;
 import io.perfmark.PerfMark;
 import java.io.ByteArrayInputStream;
-import java.io.File;
 import java.io.IOException;
 import java.io.InputStream;
 import java.net.InetSocketAddress;
@@ -90,7 +91,6 @@ import java.util.Arrays;
 import java.util.Collections;
 import java.util.LinkedList;
 import java.util.List;
-import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.CyclicBarrier;
 import java.util.concurrent.Executor;
 import java.util.concurrent.ScheduledExecutorService;
@@ -139,6 +139,7 @@ public class ServerImplTest {
       };
   private static final String AUTHORITY = "some_authority";
 
+  @SuppressWarnings("deprecation") // https://github.com/grpc/grpc-java/issues/7467
   @Rule public final ExpectedException thrown = ExpectedException.none();
 
   @BeforeClass
@@ -164,7 +165,7 @@ public class ServerImplTest {
     };
   @Mock
   private ObjectPool<Executor> executorPool;
-  private Builder builder = new Builder();
+  private ServerImplBuilder builder;
   private MutableHandlerRegistry mutableFallbackRegistry = new MutableHandlerRegistry();
   private HandlerRegistry fallbackRegistry = mock(
       HandlerRegistry.class,
@@ -201,6 +202,14 @@ public class ServerImplTest {
   @Before
   public void startUp() throws IOException {
     MockitoAnnotations.initMocks(this);
+    builder = new ServerImplBuilder(
+        new ClientTransportServersBuilder() {
+          @Override
+          public InternalServer buildClientTransportServers(
+              List<? extends ServerStreamTracer.Factory> streamTracerFactories) {
+            throw new UnsupportedOperationException();
+          }
+        });
     builder.channelz = channelz;
     builder.ticker = timer.getDeadlineTicker();
     streamTracerFactories = Arrays.asList(streamTracerFactory);
@@ -217,39 +226,19 @@ public class ServerImplTest {
   }
 
   @Test
-  public void multiport() throws Exception {
-    final CountDownLatch starts = new CountDownLatch(2);
-    final CountDownLatch shutdowns = new CountDownLatch(2);
-
-    final class Serv extends SimpleServer {
+  public void getListenSockets() throws Exception {
+    int port = 800;
+    final List<InetSocketAddress> addresses =
+        Collections.singletonList(new InetSocketAddress(800));
+    transportServer = new SimpleServer() {
       @Override
-      public void start(ServerListener listener) throws IOException {
-        super.start(listener);
-        starts.countDown();
+      public List<InetSocketAddress> getListenSocketAddresses() {
+        return addresses;
       }
-
-      @Override
-      public void shutdown() {
-        super.shutdown();
-        shutdowns.countDown();
-      }
-    }
-
-    SimpleServer transportServer1 = new Serv();
-    SimpleServer transportServer2 = new Serv();
-    assertNull(server);
-    builder.fallbackHandlerRegistry(fallbackRegistry);
-    builder.executorPool = executorPool;
-    server = new ServerImpl(
-        builder, ImmutableList.of(transportServer1, transportServer2), SERVER_CONTEXT);
-
-    server.start();
-    assertTrue(starts.await(1, TimeUnit.SECONDS));
-    assertEquals(2, shutdowns.getCount());
-
-    server.shutdown();
-    assertTrue(shutdowns.await(1, TimeUnit.SECONDS));
-    assertTrue(server.awaitTermination(1, TimeUnit.SECONDS));
+    };
+    createAndStartServer();
+    assertEquals(port, server.getPort());
+    assertThat(server.getListenSockets()).isEqualTo(addresses);
   }
 
   @Test
@@ -471,6 +460,127 @@ public class ServerImplTest {
     assertEquals(Status.Code.UNIMPLEMENTED, statusCaptor.getValue().getCode());
   }
 
+
+  @Test
+  public void executorSupplierSameExecutorBasic() throws Exception {
+    builder.executorSupplier = new ServerCallExecutorSupplier() {
+      @Override
+      public <ReqT, RespT> Executor getExecutor(ServerCall<ReqT, RespT> call, Metadata metadata) {
+        return executor.getScheduledExecutorService();
+      }
+    };
+    basicExchangeSuccessful();
+  }
+
+  @Test
+  public void executorSupplierNullBasic() throws Exception {
+    builder.executorSupplier = new ServerCallExecutorSupplier() {
+      @Override
+      public <ReqT, RespT> Executor getExecutor(ServerCall<ReqT, RespT> call, Metadata metadata) {
+        return null;
+      }
+    };
+    basicExchangeSuccessful();
+  }
+
+  @Test
+  @SuppressWarnings("unchecked")
+  public void executorSupplierSwitchExecutor() throws Exception {
+    SingleExecutor switchingExecutor = new SingleExecutor();
+    ServerCallExecutorSupplier mockSupplier = mock(ServerCallExecutorSupplier.class);
+    when(mockSupplier.getExecutor(any(ServerCall.class), any(Metadata.class)))
+            .thenReturn(switchingExecutor);
+    builder.executorSupplier = mockSupplier;
+    final AtomicReference<ServerCall<String, Integer>> callReference
+            = new AtomicReference<>();
+    mutableFallbackRegistry.addService(ServerServiceDefinition.builder(
+            new ServiceDescriptor("Waiter", METHOD))
+            .addMethod(METHOD,
+                new ServerCallHandler<String, Integer>() {
+                  @Override
+                  public ServerCall.Listener<String> startCall(
+                          ServerCall<String, Integer> call,
+                          Metadata headers) {
+                    callReference.set(call);
+                    return callListener;
+                  }
+                }).build());
+
+    createAndStartServer();
+    ServerTransportListener transportListener
+            = transportServer.registerNewServerTransport(new SimpleServerTransport());
+    transportListener.transportReady(Attributes.EMPTY);
+    Metadata requestHeaders = new Metadata();
+    StatsTraceContext statsTraceCtx =
+            StatsTraceContext.newServerContext(
+                    streamTracerFactories, "Waiter/serve", requestHeaders);
+    when(stream.statsTraceContext()).thenReturn(statsTraceCtx);
+    transportListener.streamCreated(stream, "Waiter/serve", requestHeaders);
+    verify(stream).setListener(isA(ServerStreamListener.class));
+    verify(stream, atLeast(1)).statsTraceContext();
+
+    assertEquals(1, executor.runDueTasks());
+    verify(fallbackRegistry).lookupMethod("Waiter/serve", AUTHORITY);
+    verify(streamTracerFactory).newServerStreamTracer(eq("Waiter/serve"), same(requestHeaders));
+    ArgumentCaptor<ServerCall<?,?>> callCapture = ArgumentCaptor.forClass(ServerCall.class);
+    verify(mockSupplier).getExecutor(callCapture.capture(), eq(requestHeaders));
+
+    assertThat(switchingExecutor.runnable).isNotNull();
+    assertEquals(0, executor.numPendingTasks());
+    switchingExecutor.drain();
+    ServerCall<String, Integer> call = callReference.get();
+    assertNotNull(call);
+    assertThat(call).isEqualTo(callCapture.getValue());
+  }
+
+  @Test
+  @SuppressWarnings("CheckReturnValue")
+  public void executorSupplierFutureNotSet() throws Exception {
+    builder.executorSupplier = new ServerCallExecutorSupplier() {
+      @Override
+      public <ReqT, RespT> Executor getExecutor(ServerCall<ReqT, RespT> call, Metadata metadata) {
+        throw new IllegalStateException("Yeah!");
+      }
+    };
+    doThrow(new IllegalStateException("Yeah")).doNothing()
+            .when(stream).close(any(Status.class), any(Metadata.class));
+    final AtomicReference<ServerCall<String, Integer>> callReference
+            = new AtomicReference<>();
+    mutableFallbackRegistry.addService(ServerServiceDefinition.builder(
+        new ServiceDescriptor("Waiter", METHOD))
+        .addMethod(METHOD,
+            new ServerCallHandler<String, Integer>() {
+                @Override
+                public ServerCall.Listener<String> startCall(
+                        ServerCall<String, Integer> call,
+                        Metadata headers) {
+                  callReference.set(call);
+                  return callListener;
+                }
+            }).build());
+
+    createAndStartServer();
+    ServerTransportListener transportListener
+            = transportServer.registerNewServerTransport(new SimpleServerTransport());
+    transportListener.transportReady(Attributes.EMPTY);
+    Metadata requestHeaders = new Metadata();
+    StatsTraceContext statsTraceCtx =
+            StatsTraceContext.newServerContext(
+                    streamTracerFactories, "Waiter/serve", requestHeaders);
+    when(stream.statsTraceContext()).thenReturn(statsTraceCtx);
+    transportListener.streamCreated(stream, "Waiter/serve", requestHeaders);
+    verify(stream).setListener(isA(ServerStreamListener.class));
+    verify(stream, atLeast(1)).statsTraceContext();
+
+    assertEquals(1, executor.runDueTasks());
+    verify(fallbackRegistry).lookupMethod("Waiter/serve", AUTHORITY);
+    assertThat(callReference.get()).isNull();
+    verify(stream, times(2)).close(statusCaptor.capture(), any(Metadata.class));
+    Status status = statusCaptor.getAllValues().get(1);
+    assertEquals(Code.UNKNOWN, status.getCode());
+    assertThat(status.getCause() instanceof IllegalStateException);
+  }
+
   @Test
   public void decompressorNotFound() throws Exception {
     String decompressorName = "NON_EXISTENT_DECOMPRESSOR";
@@ -487,6 +597,7 @@ public class ServerImplTest {
 
     transportListener.streamCreated(stream, "Waiter/nonexist", requestHeaders);
 
+    verify(stream).setListener(isA(ServerStreamListener.class));
     verify(stream).streamId();
     verify(stream).close(statusCaptor.capture(), any(Metadata.class));
     Status status = statusCaptor.getValue();
@@ -560,6 +671,7 @@ public class ServerImplTest {
     Context callContext = callContextReference.get();
     assertNotNull(callContext);
     assertEquals("context added by tracer", SERVER_TRACER_ADDED_KEY.get(callContext));
+    assertEquals(server, io.grpc.InternalServer.SERVER_CONTEXT_KEY.get(callContext));
 
     streamListener.messagesAvailable(new SingleMessageProducer(STRING_MARSHALLER.stream(request)));
     assertEquals(1, executor.runDueTasks());
@@ -1084,7 +1196,9 @@ public class ServerImplTest {
         context, contextCancelled, null);
 
     // For close status OK:
-    // isCancelled is expected to be true after all pending work is done
+    // The context isCancelled is expected to be true after all pending work is done,
+    // but for the call it should be false as it gets set cancelled only if the call
+    // fails with a non-OK status.
     assertFalse(callReference.get().isCancelled());
     assertFalse(context.get().isCancelled());
     streamListener.closed(Status.OK);
@@ -1092,7 +1206,7 @@ public class ServerImplTest {
     assertFalse(context.get().isCancelled());
 
     assertEquals(1, executor.runDueTasks());
-    assertTrue(callReference.get().isCancelled());
+    assertFalse(callReference.get().isCancelled());
     assertTrue(context.get().isCancelled());
     assertTrue(contextCancelled.get());
   }
@@ -1120,15 +1234,22 @@ public class ServerImplTest {
   @Test
   public void getPort() throws Exception {
     final InetSocketAddress addr = new InetSocketAddress(65535);
+    final List<InetSocketAddress> addrs = Collections.singletonList(addr);
     transportServer = new SimpleServer() {
       @Override
-      public SocketAddress getListenSocketAddress() {
+      public InetSocketAddress getListenSocketAddress() {
         return addr;
       }
+
+      @Override
+      public List<InetSocketAddress> getListenSocketAddresses() {
+        return addrs;
+      }
     };
     createAndStartServer();
 
     assertThat(server.getPort()).isEqualTo(addr.getPort());
+    assertThat(server.getListenSockets()).isEqualTo(addrs);
   }
 
   @Test
@@ -1420,7 +1541,7 @@ public class ServerImplTest {
 
     builder.fallbackHandlerRegistry(fallbackRegistry);
     builder.executorPool = executorPool;
-    server = new ServerImpl(builder, Collections.singletonList(transportServer), SERVER_CONTEXT);
+    server = new ServerImpl(builder, transportServer, SERVER_CONTEXT);
   }
 
   private void verifyExecutorsAcquired() {
@@ -1439,8 +1560,9 @@ public class ServerImplTest {
 
   private void ensureServerStateNotLeaked() {
     verify(stream).close(statusCaptor.capture(), metadataCaptor.capture());
-    assertEquals(Status.UNKNOWN, statusCaptor.getValue());
-    assertNull(statusCaptor.getValue().getCause());
+    assertEquals(Status.UNKNOWN.getCode(), statusCaptor.getValue().getCode());
+    // Used in InProcessTransport when set to include the cause with the status
+    assertNotNull(statusCaptor.getValue().getCause());
     assertTrue(metadataCaptor.getValue().keys().isEmpty());
   }
 
@@ -1457,11 +1579,21 @@ public class ServerImplTest {
       return new InetSocketAddress(12345);
     }
 
+    @Override
+    public List<InetSocketAddress> getListenSocketAddresses() {
+      return Collections.singletonList(new InetSocketAddress(12345));
+    }
+
     @Override
     public InternalInstrumented<SocketStats> getListenSocketStats() {
       return null;
     }
 
+    @Override
+    public List<InternalInstrumented<SocketStats>> getListenSocketStatsList() {
+      return null;
+    }
+
     @Override
     public void shutdown() {
       listener.serverShutdown();
@@ -1504,17 +1636,26 @@ public class ServerImplTest {
     }
   }
 
-  private static class Builder extends AbstractServerImplBuilder<Builder> {
-    @Override protected List<InternalServer> buildTransportServers(
-        List<? extends ServerStreamTracer.Factory> streamTracerFactories) {
-      throw new UnsupportedOperationException();
-    }
-
-    @Override public Builder useTransportSecurity(File f1, File f2)  {
-      throw new UnsupportedOperationException();
-    }
-  }
-
   /** Allows more precise catch blocks than plain Error to avoid catching AssertionError. */
   private static final class TestError extends Error {}
+
+  private static class SingleExecutor implements Executor {
+    private Runnable runnable;
+
+    @Override
+    public void execute(Runnable r) {
+      if (runnable != null) {
+        fail("Already have runnable scheduled");
+      }
+      runnable = r;
+    }
+
+    public void drain() {
+      if (runnable != null) {
+        Runnable r = runnable;
+        runnable = null;
+        r.run();
+      }
+    }
+  }
 }
