@@ -19,7 +19,9 @@ package io.grpc.util;
 import static com.google.common.truth.Truth.assertThat;
 import static io.grpc.ConnectivityState.CONNECTING;
 import static io.grpc.ConnectivityState.READY;
+import static io.grpc.ConnectivityState.TRANSIENT_FAILURE;
 import static io.grpc.util.GracefulSwitchLoadBalancer.BUFFER_PICKER;
+import static org.mockito.ArgumentMatchers.eq;
 import static org.mockito.Mockito.doReturn;
 import static org.mockito.Mockito.inOrder;
 import static org.mockito.Mockito.mock;
@@ -33,6 +35,7 @@ import io.grpc.EquivalentAddressGroup;
 import io.grpc.LoadBalancer;
 import io.grpc.LoadBalancer.CreateSubchannelArgs;
 import io.grpc.LoadBalancer.Helper;
+import io.grpc.LoadBalancer.PickSubchannelArgs;
 import io.grpc.LoadBalancer.ResolvedAddresses;
 import io.grpc.LoadBalancer.Subchannel;
 import io.grpc.LoadBalancer.SubchannelPicker;
@@ -40,8 +43,10 @@ import io.grpc.LoadBalancerProvider;
 import io.grpc.LoadBalancerRegistry;
 import io.grpc.Status;
 import java.net.SocketAddress;
+import java.util.ArrayList;
 import java.util.Collections;
 import java.util.HashMap;
+import java.util.List;
 import java.util.Map;
 import org.junit.Before;
 import org.junit.Rule;
@@ -49,6 +54,7 @@ import org.junit.Test;
 import org.junit.rules.ExpectedException;
 import org.junit.runner.RunWith;
 import org.junit.runners.JUnit4;
+import org.mockito.ArgumentCaptor;
 import org.mockito.InOrder;
 
 /**
@@ -56,6 +62,7 @@ import org.mockito.InOrder;
  */
 @RunWith(JUnit4.class)
 public class GracefulSwitchLoadBalancerTest {
+  @SuppressWarnings("deprecation") // https://github.com/grpc/grpc-java/issues/7467
   @Rule
   public final ExpectedException thrown = ExpectedException.none();
 
@@ -420,6 +427,65 @@ public class GracefulSwitchLoadBalancerTest {
     verifyNoMoreInteractions(lb0, lb1);
   }
 
+
+  @Test
+  public void newLbFactoryEqualToOldOneShouldHaveNoEffect() {
+    final List<LoadBalancer> balancers = new ArrayList<>();
+
+    final class LoadBalancerFactoryWithId extends LoadBalancer.Factory {
+      final int id;
+
+      LoadBalancerFactoryWithId(int id) {
+        this.id = id;
+      }
+
+      @Override
+      public LoadBalancer newLoadBalancer(Helper helper) {
+        LoadBalancer balancer = mock(LoadBalancer.class);
+        balancers.add(balancer);
+        return balancer;
+      }
+
+      @Override
+      public boolean equals(Object o) {
+        if (!(o instanceof LoadBalancerFactoryWithId)) {
+          return false;
+        }
+        LoadBalancerFactoryWithId that = (LoadBalancerFactoryWithId) o;
+        return id == that.id;
+      }
+
+      @Override
+      public int hashCode() {
+        return id;
+      }
+    }
+
+    gracefulSwitchLb.switchTo(new LoadBalancerFactoryWithId(0));
+    assertThat(balancers).hasSize(1);
+    LoadBalancer lb0 = balancers.get(0);
+
+    gracefulSwitchLb.switchTo(new LoadBalancerFactoryWithId(0));
+    assertThat(balancers).hasSize(1);
+
+    gracefulSwitchLb.switchTo(new LoadBalancerFactoryWithId(1));
+    assertThat(balancers).hasSize(2);
+    LoadBalancer lb1 = balancers.get(1);
+    verify(lb0).shutdown();
+
+    verifyNoMoreInteractions(lb0, lb1);
+  }
+
+  @Test
+  public void transientFailureOnInitialResolutionError() {
+    gracefulSwitchLb.handleNameResolutionError(Status.DATA_LOSS);
+    ArgumentCaptor<SubchannelPicker> pickerCaptor = ArgumentCaptor.forClass(null);
+    verify(mockHelper).updateBalancingState(eq(TRANSIENT_FAILURE), pickerCaptor.capture());
+    SubchannelPicker picker = pickerCaptor.getValue();
+    assertThat(picker.pickSubchannel(mock(PickSubchannelArgs.class)).getStatus().getCode())
+        .isEqualTo(Status.Code.DATA_LOSS);
+  }
+
   @Deprecated
   @Test
   public void handleSubchannelState_shouldThrow() {
