@@ -16,6 +16,7 @@
 
 package io.grpc.examples.helloworld;
 
+import io.grpc.Channel;
 import io.grpc.ManagedChannel;
 import io.grpc.ManagedChannelBuilder;
 import io.grpc.StatusRuntimeException;
@@ -29,26 +30,15 @@ import java.util.logging.Logger;
 public class HelloWorldClient {
   private static final Logger logger = Logger.getLogger(HelloWorldClient.class.getName());
 
-  private final ManagedChannel channel;
   private final GreeterGrpc.GreeterBlockingStub blockingStub;
 
-  /** Construct client connecting to HelloWorld server at {@code host:port}. */
-  public HelloWorldClient(String host, int port) {
-    this(ManagedChannelBuilder.forAddress(host, port)
-        // Channels are secure by default (via SSL/TLS). For the example we disable TLS to avoid
-        // needing certificates.
-        .usePlaintext()
-        .build());
-  }
-
   /** Construct client for accessing HelloWorld server using the existing channel. */
-  HelloWorldClient(ManagedChannel channel) {
-    this.channel = channel;
-    blockingStub = GreeterGrpc.newBlockingStub(channel);
-  }
+  public HelloWorldClient(Channel channel) {
+    // 'channel' here is a Channel, not a ManagedChannel, so it is not this code's responsibility to
+    // shut it down.
 
-  public void shutdown() throws InterruptedException {
-    channel.shutdown().awaitTermination(5, TimeUnit.SECONDS);
+    // Passing Channels to code makes code easier to test and makes it easier to reuse Channels.
+    blockingStub = GreeterGrpc.newBlockingStub(channel);
   }
 
   /** Say hello to server. */
@@ -67,20 +57,43 @@ public class HelloWorldClient {
 
   /**
    * Greet server. If provided, the first element of {@code args} is the name to use in the
-   * greeting.
+   * greeting. The second argument is the target server.
    */
   public static void main(String[] args) throws Exception {
+    String user = "world";
     // Access a service running on the local machine on port 50051
-    HelloWorldClient client = new HelloWorldClient("localhost", 50051);
-    try {
-      String user = "world";
-      // Use the arg as the name to greet if provided
-      if (args.length > 0) {
-        user = args[0];
+    String target = "localhost:50051";
+    // Allow passing in the user and target strings as command line arguments
+    if (args.length > 0) {
+      if ("--help".equals(args[0])) {
+        System.err.println("Usage: [name [target]]");
+        System.err.println("");
+        System.err.println("  name    The name you wish to be greeted by. Defaults to " + user);
+        System.err.println("  target  The server to connect to. Defaults to " + target);
+        System.exit(1);
       }
+      user = args[0];
+    }
+    if (args.length > 1) {
+      target = args[1];
+    }
+
+    // Create a communication channel to the server, known as a Channel. Channels are thread-safe
+    // and reusable. It is common to create channels at the beginning of your application and reuse
+    // them until the application shuts down.
+    ManagedChannel channel = ManagedChannelBuilder.forTarget(target)
+        // Channels are secure by default (via SSL/TLS). For the example we disable TLS to avoid
+        // needing certificates.
+        .usePlaintext()
+        .build();
+    try {
+      HelloWorldClient client = new HelloWorldClient(channel);
       client.greet(user);
     } finally {
-      client.shutdown();
+      // ManagedChannels use resources like threads and TCP connections. To prevent leaking these
+      // resources the channel should be shut down when it will no longer be used. If it may be used
+      // again leave it running.
+      channel.shutdownNow().awaitTermination(5, TimeUnit.SECONDS);
     }
   }
 }
