@@ -40,7 +40,6 @@ import java.util.Random;
 import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicReference;
-import org.junit.After;
 import org.junit.Before;
 import org.junit.Rule;
 import org.junit.Test;
@@ -53,6 +52,10 @@ import org.mockito.ArgumentCaptor;
  * For demonstrating how to write gRPC unit test only.
  * Not intended to provide a high code coverage or to test every major usecase.
  *
+ * directExecutor() makes it easier to have deterministic tests.
+ * However, if your implementation uses another thread and uses streaming it is better to use
+ * the default executor, to avoid hitting bug #3084.
+ *
  * <p>For basic unit test examples see {@link io.grpc.examples.helloworld.HelloWorldClientTest} and
  * {@link io.grpc.examples.helloworld.HelloWorldServerTest}.
  */
@@ -92,16 +95,11 @@ public class RouteGuideClientTest {
     // Use a mutable service registry for later registering the service impl for each test case.
     grpcCleanup.register(InProcessServerBuilder.forName(serverName)
         .fallbackHandlerRegistry(serviceRegistry).directExecutor().build().start());
-    client =
-        new RouteGuideClient(InProcessChannelBuilder.forName(serverName).directExecutor());
+    client = new RouteGuideClient(grpcCleanup.register(
+        InProcessChannelBuilder.forName(serverName).directExecutor().build()));
     client.setTestHelper(testHelper);
   }
 
-  @After
-  public void tearDown() throws Exception {
-    client.shutdown();
-  }
-
   /**
    * Example for testing blocking unary call.
    */
@@ -402,9 +400,9 @@ public class RouteGuideClientTest {
     assertEquals(
         Arrays.asList(
             Point.newBuilder().setLatitude(0).setLongitude(0).build(),
-            Point.newBuilder().setLatitude(0).setLongitude(1).build(),
-            Point.newBuilder().setLatitude(1).setLongitude(0).build(),
-            Point.newBuilder().setLatitude(1).setLongitude(1).build()
+            Point.newBuilder().setLatitude(0).setLongitude(10_000_000).build(),
+            Point.newBuilder().setLatitude(10_000_000).setLongitude(0).build(),
+            Point.newBuilder().setLatitude(10_000_000).setLongitude(10_000_000).build()
         ),
         locationsDelivered);
 
