@@ -33,6 +33,7 @@ import com.google.common.util.concurrent.ListenableFuture;
 import com.google.common.util.concurrent.ListeningExecutorService;
 import com.google.common.util.concurrent.MoreExecutors;
 import io.grpc.ManagedChannel;
+import io.grpc.ManagedChannelBuilder;
 import io.grpc.Server;
 import io.grpc.ServerBuilder;
 import io.grpc.Status;
@@ -515,6 +516,11 @@ public class StressTestClient {
         return Worker.this.channel;
       }
 
+      @Override
+      protected ManagedChannelBuilder<?> createChannelBuilder() {
+        throw new UnsupportedOperationException();
+      }
+
       @Override
       protected int operationTimeoutMillis() {
         // Don't enforce a timeout when using the interop tests for the stress test client.
