@@ -54,7 +54,9 @@ public enum TestCases {
   CANCEL_AFTER_FIRST_RESPONSE("cancel on first response"),
   TIMEOUT_ON_SLEEPING_SERVER("timeout before receiving a response"),
   VERY_LARGE_REQUEST("very large request"),
-  PICK_FIRST_UNARY("all requests are sent to one server despite multiple servers are resolved");
+  PICK_FIRST_UNARY("all requests are sent to one server despite multiple servers are resolved"),
+  RPC_SOAK("sends 'soak_iterations' large_unary rpcs in a loop, each on the same channel"),
+  CHANNEL_SOAK("sends 'soak_iterations' large_unary rpcs in a loop, each on a new channel");
 
   private final String description;
 
