@@ -18,23 +18,20 @@ package io.grpc.testing.integration;
 
 import com.google.common.base.Preconditions;
 import com.google.common.util.concurrent.MoreExecutors;
+import io.grpc.ChannelCredentials;
+import io.grpc.Grpc;
 import io.grpc.ManagedChannel;
 import io.grpc.Server;
+import io.grpc.ServerCredentials;
+import io.grpc.TlsChannelCredentials;
+import io.grpc.TlsServerCredentials;
 import io.grpc.internal.testing.TestUtils;
-import io.grpc.netty.GrpcSslContexts;
-import io.grpc.netty.NegotiationType;
-import io.grpc.netty.NettyChannelBuilder;
-import io.grpc.netty.NettyServerBuilder;
 import io.grpc.stub.StreamObserver;
 import io.grpc.testing.integration.Messages.ResponseParameters;
 import io.grpc.testing.integration.Messages.StreamingOutputCallRequest;
 import io.grpc.testing.integration.Messages.StreamingOutputCallResponse;
-import io.netty.handler.ssl.ClientAuth;
-import io.netty.handler.ssl.SslContext;
 import java.io.File;
 import java.io.IOException;
-import java.security.cert.CertificateException;
-import java.security.cert.X509Certificate;
 import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.CyclicBarrier;
 import java.util.concurrent.ExecutorService;
@@ -184,43 +181,35 @@ public class ConcurrencyTest {
   /**
    * Creates and starts a new {@link TestServiceImpl} server.
    */
-  private Server newServer() throws CertificateException, IOException {
+  private Server newServer() throws IOException {
     File serverCertChainFile = TestUtils.loadCert("server1.pem");
     File serverPrivateKeyFile = TestUtils.loadCert("server1.key");
-    X509Certificate[] serverTrustedCaCerts = {
-      TestUtils.loadX509Cert("ca.pem")
-    };
+    File serverTrustedCaCerts = TestUtils.loadCert("ca.pem");
 
-    SslContext sslContext =
-        GrpcSslContexts.forServer(serverCertChainFile, serverPrivateKeyFile)
-                       .trustManager(serverTrustedCaCerts)
-                       .clientAuth(ClientAuth.REQUIRE)
-                       .build();
+    ServerCredentials serverCreds = TlsServerCredentials.newBuilder()
+        .keyManager(serverCertChainFile, serverPrivateKeyFile)
+        .trustManager(serverTrustedCaCerts)
+        .clientAuth(TlsServerCredentials.ClientAuth.REQUIRE)
+        .build();
 
-    return NettyServerBuilder.forPort(0)
-        .sslContext(sslContext)
+    return Grpc.newServerBuilderForPort(0, serverCreds)
         .addService(new TestServiceImpl(serverExecutor))
         .build()
         .start();
   }
 
-  private ManagedChannel newClientChannel() throws CertificateException, IOException {
+  private ManagedChannel newClientChannel() throws IOException {
     File clientCertChainFile = TestUtils.loadCert("client.pem");
     File clientPrivateKeyFile = TestUtils.loadCert("client.key");
-    X509Certificate[] clientTrustedCaCerts = {
-      TestUtils.loadX509Cert("ca.pem")
-    };
+    File clientTrustedCaCerts = TestUtils.loadCert("ca.pem");
 
-    SslContext sslContext =
-        GrpcSslContexts.forClient()
-                       .keyManager(clientCertChainFile, clientPrivateKeyFile)
-                       .trustManager(clientTrustedCaCerts)
-                       .build();
+    ChannelCredentials channelCreds = TlsChannelCredentials.newBuilder()
+        .keyManager(clientCertChainFile, clientPrivateKeyFile)
+        .trustManager(clientTrustedCaCerts)
+        .build();
 
-    return NettyChannelBuilder.forAddress("localhost", server.getPort())
+    return Grpc.newChannelBuilder("localhost:" + server.getPort(), channelCreds)
         .overrideAuthority(TestUtils.TEST_SERVER_HOST)
-        .negotiationType(NegotiationType.TLS)
-        .sslContext(sslContext)
         .build();
   }
 }
