@@ -73,7 +73,9 @@ public class TestCasesTest {
       "client_compressed_unary_noprobe",
       "client_compressed_streaming_noprobe",
       "very_large_request",
-      "pick_first_unary"
+      "pick_first_unary",
+      "channel_soak",
+      "rpc_soak"
     };
 
     assertEquals(testCases.length + additionalTestCases.length, TestCases.values().length);
