@@ -25,7 +25,7 @@ import org.junit.runners.JUnit4;
 public class TestServiceClientTest {
 
   @Test
-  public void emptyArgumentListShouldNotThrowException() {
+  public void emptyArgumentListShouldNotThrowException() throws Exception {
     TestServiceClient client = new TestServiceClient();
     client.parseArgs(new String[0]);
     client.setUp();
