@@ -17,10 +17,11 @@
 package io.grpc.netty;
 
 import io.grpc.Internal;
+import io.grpc.ServerCredentials;
 import io.grpc.ServerProvider;
+import java.net.InetSocketAddress;
 
-
-/** Provider for {@link NettyChannelBuilder} instances. */
+/** Provider for {@link NettyServerBuilder} instances. */
 @Internal
 public final class NettyServerProvider extends ServerProvider {
 
@@ -38,5 +39,15 @@ public final class NettyServerProvider extends ServerProvider {
   protected NettyServerBuilder builderForPort(int port) {
     return NettyServerBuilder.forPort(port);
   }
+
+  @Override
+  protected NewServerBuilderResult newServerBuilderForPort(int port, ServerCredentials creds) {
+    ProtocolNegotiators.FromServerCredentialsResult result = ProtocolNegotiators.from(creds);
+    if (result.error != null) {
+      return NewServerBuilderResult.error(result.error);
+    }
+    return NewServerBuilderResult.serverBuilder(
+        new NettyServerBuilder(new InetSocketAddress(port), result.negotiator));
+  }
 }
 
