@@ -32,6 +32,7 @@ import io.grpc.Metadata;
 import io.grpc.Status;
 import io.grpc.internal.GrpcUtil;
 import io.grpc.internal.SharedResourceHolder.Resource;
+import io.grpc.internal.TransportTracer;
 import io.grpc.netty.GrpcHttp2HeadersUtils.GrpcHttp2InboundHeaders;
 import io.grpc.netty.NettySocketSupport.NativeSocketOptions;
 import io.netty.buffer.ByteBufAllocator;
@@ -46,22 +47,27 @@ import io.netty.channel.ServerChannel;
 import io.netty.channel.nio.NioEventLoopGroup;
 import io.netty.channel.socket.nio.NioServerSocketChannel;
 import io.netty.channel.socket.nio.NioSocketChannel;
+import io.netty.handler.codec.DecoderException;
+import io.netty.handler.codec.http2.Http2Connection;
 import io.netty.handler.codec.http2.Http2Exception;
+import io.netty.handler.codec.http2.Http2FlowController;
 import io.netty.handler.codec.http2.Http2Headers;
+import io.netty.handler.codec.http2.Http2Stream;
 import io.netty.util.AsciiString;
 import io.netty.util.NettyRuntime;
 import io.netty.util.concurrent.DefaultThreadFactory;
 import java.io.IOException;
 import java.lang.reflect.Constructor;
 import java.nio.channels.ClosedChannelException;
+import java.nio.channels.UnresolvedAddressException;
 import java.util.Map;
-import java.util.Map.Entry;
 import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.TimeUnit;
 import java.util.logging.Level;
 import java.util.logging.Logger;
 import javax.annotation.CheckReturnValue;
 import javax.annotation.Nullable;
+import javax.net.ssl.SSLException;
 
 /**
  * Common utility methods.
@@ -86,41 +92,15 @@ class Utils {
   public static final Resource<EventLoopGroup> DEFAULT_BOSS_EVENT_LOOP_GROUP;
   public static final Resource<EventLoopGroup> DEFAULT_WORKER_EVENT_LOOP_GROUP;
 
-  public static final Resource<ByteBufAllocator> BYTE_BUF_ALLOCATOR =
-      new Resource<ByteBufAllocator>() {
-        @Override
-        public ByteBufAllocator create() {
-          if (Boolean.parseBoolean(
-                  System.getProperty("io.grpc.netty.useCustomAllocator", "false"))) {
-            int maxOrder;
-            if (System.getProperty("io.netty.allocator.maxOrder") == null) {
-              // See the implementation of PooledByteBufAllocator.  DEFAULT_MAX_ORDER in there is
-              // 11, which makes chunk size to be 8192 << 11 = 16 MiB.  We want the chunk size to be
-              // 2MiB, thus reducing the maxOrder to 8.
-              maxOrder = 8;
-            } else {
-              maxOrder = PooledByteBufAllocator.defaultMaxOrder();
-            }
-            return new PooledByteBufAllocator(
-                PooledByteBufAllocator.defaultPreferDirect(),
-                PooledByteBufAllocator.defaultNumHeapArena(),
-                PooledByteBufAllocator.defaultNumDirectArena(),
-                PooledByteBufAllocator.defaultPageSize(),
-                maxOrder,
-                PooledByteBufAllocator.defaultTinyCacheSize(),
-                PooledByteBufAllocator.defaultSmallCacheSize(),
-                PooledByteBufAllocator.defaultNormalCacheSize(),
-                PooledByteBufAllocator.defaultUseCacheForAllThreads());
-          } else {
-            return ByteBufAllocator.DEFAULT;
-          }
-        }
+  // This class is initialized on first use, thus provides delayed allocator creation.
+  private static final class ByteBufAllocatorPreferDirectHolder {
+    private static final ByteBufAllocator allocator = createByteBufAllocator(true);
+  }
 
-        @Override
-        public void close(ByteBufAllocator allocator) {
-          // PooledByteBufAllocator doesn't provide a shutdown method.  Leaving it to GC.
-        }
-      };
+  // This class is initialized on first use, thus provides delayed allocator creation.
+  private static final class ByteBufAllocatorPreferHeapHolder {
+    private static final ByteBufAllocator allocator = createByteBufAllocator(false);
+  }
 
   public static final ChannelFactory<? extends ServerChannel> DEFAULT_SERVER_CHANNEL_FACTORY;
   public static final Class<? extends Channel> DEFAULT_CLIENT_CHANNEL_TYPE;
@@ -148,6 +128,54 @@ class Utils {
     }
   }
 
+  public static ByteBufAllocator getByteBufAllocator(boolean forceHeapBuffer) {
+    if (Boolean.parseBoolean(
+            System.getProperty("io.grpc.netty.useCustomAllocator", "true"))) {
+      boolean defaultPreferDirect = PooledByteBufAllocator.defaultPreferDirect();
+      logger.log(
+          Level.FINE,
+          String.format(
+              "Using custom allocator: forceHeapBuffer=%s, defaultPreferDirect=%s",
+              forceHeapBuffer,
+              defaultPreferDirect));
+      if (forceHeapBuffer || !defaultPreferDirect) {
+        return ByteBufAllocatorPreferHeapHolder.allocator;
+      } else {
+        return ByteBufAllocatorPreferDirectHolder.allocator;
+      }
+    } else {
+      logger.log(Level.FINE, "Using default allocator");
+      return ByteBufAllocator.DEFAULT;
+    }
+  }
+
+  private static ByteBufAllocator createByteBufAllocator(boolean preferDirect) {
+    int maxOrder;
+    logger.log(Level.FINE, "Creating allocator, preferDirect=" + preferDirect);
+    if (System.getProperty("io.netty.allocator.maxOrder") == null) {
+      // See the implementation of PooledByteBufAllocator.  DEFAULT_MAX_ORDER in there is
+      // 11, which makes chunk size to be 8192 << 11 = 16 MiB.  We want the chunk size to be
+      // 2MiB, thus reducing the maxOrder to 8.
+      maxOrder = 8;
+      logger.log(Level.FINE, "Forcing maxOrder=" + maxOrder);
+    } else {
+      maxOrder = PooledByteBufAllocator.defaultMaxOrder();
+      logger.log(Level.FINE, "Using default maxOrder=" + maxOrder);
+    }
+    return new PooledByteBufAllocator(
+        preferDirect,
+        PooledByteBufAllocator.defaultNumHeapArena(),
+        // Assuming neither gRPC nor netty are using allocator.directBuffer() to request
+        // specifically for direct buffers, which is true as I just checked, setting arenas to 0
+        // will make sure no direct buffer is ever created.
+        preferDirect ? PooledByteBufAllocator.defaultNumDirectArena() : 0,
+        PooledByteBufAllocator.defaultPageSize(),
+        maxOrder,
+        PooledByteBufAllocator.defaultSmallCacheSize(),
+        PooledByteBufAllocator.defaultNormalCacheSize(),
+        PooledByteBufAllocator.defaultUseCacheForAllThreads());
+  }
+
   public static Metadata convertHeaders(Http2Headers http2Headers) {
     if (http2Headers instanceof GrpcHttp2InboundHeaders) {
       GrpcHttp2InboundHeaders h = (GrpcHttp2InboundHeaders) http2Headers;
@@ -244,9 +272,15 @@ class Utils {
       extraT.initCause(t);
       return Status.UNKNOWN.withDescription("channel closed").withCause(extraT);
     }
+    if (t instanceof DecoderException && t.getCause() instanceof SSLException) {
+      return Status.UNAVAILABLE.withDescription("ssl exception").withCause(t);
+    }
     if (t instanceof IOException) {
       return Status.UNAVAILABLE.withDescription("io exception").withCause(t);
     }
+    if (t instanceof UnresolvedAddressException) {
+      return Status.UNAVAILABLE.withDescription("unresolved address").withCause(t);
+    }
     if (t instanceof Http2Exception) {
       return Status.INTERNAL.withDescription("http2 exception").withCause(t);
     }
@@ -411,6 +445,25 @@ class Utils {
     }
   }
 
+  static final class FlowControlReader implements TransportTracer.FlowControlReader {
+    private final Http2Stream connectionStream;
+    private final Http2FlowController local;
+    private final Http2FlowController remote;
+
+    FlowControlReader(Http2Connection connection) {
+      local = connection.local().flowController();
+      remote = connection.remote().flowController();
+      connectionStream = connection.connectionStream();
+    }
+
+    @Override
+    public TransportTracer.FlowControlWindows read() {
+      return new TransportTracer.FlowControlWindows(
+          local.windowSize(connectionStream),
+          remote.windowSize(connectionStream));
+    }
+  }
+
   static InternalChannelz.SocketOptions getSocketOptions(Channel channel) {
     ChannelConfig config = channel.config();
     InternalChannelz.SocketOptions.Builder b = new InternalChannelz.SocketOptions.Builder();
@@ -428,7 +481,7 @@ class Utils {
       b.setSocketOptionTimeoutMillis(timeoutMillis);
     }
 
-    for (Entry<ChannelOption<?>, Object> opt : config.getOptions().entrySet()) {
+    for (Map.Entry<ChannelOption<?>, Object> opt : config.getOptions().entrySet()) {
       ChannelOption<?> key = opt.getKey();
       // Constants are pooled, so there should only be one instance of each constant
       if (key.equals(SO_LINGER) || key.equals(SO_TIMEOUT)) {
@@ -443,7 +496,7 @@ class Utils {
         = NettySocketSupport.getNativeSocketOptions(channel);
     if (nativeOptions != null) {
       b.setTcpInfo(nativeOptions.tcpInfo); // may be null
-      for (Entry<String, String> entry : nativeOptions.otherInfo.entrySet()) {
+      for (Map.Entry<String, String> entry : nativeOptions.otherInfo.entrySet()) {
         b.addOption(entry.getKey(), entry.getValue());
       }
     }
