@@ -16,12 +16,19 @@
 
 package io.grpc.netty;
 
+import static com.google.common.truth.Truth.assertThat;
 import static org.junit.Assert.assertEquals;
+import static org.junit.Assert.assertNotNull;
 import static org.junit.Assert.assertTrue;
 import static org.mockito.Mockito.mock;
 
+import io.grpc.ChannelCredentials;
+import io.grpc.InsecureChannelCredentials;
 import io.grpc.ManagedChannel;
-import io.grpc.netty.InternalNettyChannelBuilder.OverrideAuthorityChecker;
+import io.grpc.internal.ClientTransportFactory;
+import io.grpc.internal.ClientTransportFactory.SwapChannelCredentialsResult;
+import io.grpc.netty.NettyTestUtil.TrackingObjectPoolForTest;
+import io.grpc.netty.ProtocolNegotiators.PlaintextProtocolNegotiatorClientFactory;
 import io.netty.channel.Channel;
 import io.netty.channel.ChannelFactory;
 import io.netty.channel.EventLoopGroup;
@@ -40,9 +47,10 @@ import org.junit.runners.JUnit4;
 @RunWith(JUnit4.class)
 public class NettyChannelBuilderTest {
 
+  @SuppressWarnings("deprecation") // https://github.com/grpc/grpc-java/issues/7467
   @Rule public final ExpectedException thrown = ExpectedException.none();
   private final SslContext noSslContext = null;
-  
+
   private void shutdown(ManagedChannel mc) throws Exception {
     mc.shutdownNow();
     assertTrue(mc.awaitTermination(1, TimeUnit.SECONDS));
@@ -89,20 +97,6 @@ public class NettyChannelBuilderTest {
     }
   }
 
-  @Test
-  public void overrideAllowsInvalidAuthority() {
-    NettyChannelBuilder builder = new NettyChannelBuilder(new SocketAddress(){});
-    InternalNettyChannelBuilder.overrideAuthorityChecker(builder, new OverrideAuthorityChecker() {
-      @Override
-      public String checkAuthority(String authority) {
-        return authority;
-      }
-    });
-    Object unused = builder.overrideAuthority("[invalidauthority")
-        .negotiationType(NegotiationType.PLAINTEXT)
-        .buildTransportFactory();
-  }
-
   @Test
   public void failOverrideInvalidAuthority() {
     NettyChannelBuilder builder = new NettyChannelBuilder(new SocketAddress(){});
@@ -113,6 +107,27 @@ public class NettyChannelBuilderTest {
     builder.overrideAuthority("[invalidauthority");
   }
 
+  @Test
+  public void disableCheckAuthorityAllowsInvalidAuthority() {
+    NettyChannelBuilder builder = new NettyChannelBuilder(new SocketAddress(){})
+        .disableCheckAuthority();
+
+    Object unused = builder.overrideAuthority("[invalidauthority")
+        .negotiationType(NegotiationType.PLAINTEXT)
+        .buildTransportFactory();
+  }
+
+  @Test
+  public void enableCheckAuthorityFailOverrideInvalidAuthority() {
+    NettyChannelBuilder builder = new NettyChannelBuilder(new SocketAddress(){})
+        .disableCheckAuthority()
+        .enableCheckAuthority();
+
+    thrown.expect(IllegalArgumentException.class);
+    thrown.expectMessage("Invalid authority:");
+    builder.overrideAuthority("[invalidauthority");
+  }
+
   @Test
   public void failInvalidAuthority() {
     thrown.expect(IllegalArgumentException.class);
@@ -139,22 +154,46 @@ public class NettyChannelBuilderTest {
     builder.sslContext(sslContext);
   }
 
+  @Test
+  public void failNegotiationTypeWithChannelCredentials_target() {
+    NettyChannelBuilder builder = NettyChannelBuilder.forTarget(
+        "fakeTarget", InsecureChannelCredentials.create());
+
+    thrown.expect(IllegalStateException.class);
+    thrown.expectMessage("Cannot change security when using ChannelCredentials");
+
+    builder.negotiationType(NegotiationType.TLS);
+  }
+
+  @Test
+  public void failNegotiationTypeWithChannelCredentials_socketAddress() {
+    NettyChannelBuilder builder = NettyChannelBuilder.forAddress(
+        new SocketAddress(){}, InsecureChannelCredentials.create());
+
+    thrown.expect(IllegalStateException.class);
+    thrown.expectMessage("Cannot change security when using ChannelCredentials");
+
+    builder.negotiationType(NegotiationType.TLS);
+  }
+
   @Test
   public void createProtocolNegotiatorByType_plaintext() {
     ProtocolNegotiator negotiator = NettyChannelBuilder.createProtocolNegotiatorByType(
         NegotiationType.PLAINTEXT,
-        noSslContext);
+        noSslContext, null);
     // just check that the classes are the same, and that negotiator is not null.
     assertTrue(negotiator instanceof ProtocolNegotiators.PlaintextProtocolNegotiator);
+    negotiator.close();
   }
 
   @Test
   public void createProtocolNegotiatorByType_plaintextUpgrade() {
     ProtocolNegotiator negotiator = NettyChannelBuilder.createProtocolNegotiatorByType(
         NegotiationType.PLAINTEXT_UPGRADE,
-        noSslContext);
+        noSslContext, null);
     // just check that the classes are the same, and that negotiator is not null.
     assertTrue(negotiator instanceof ProtocolNegotiators.PlaintextUpgradeProtocolNegotiator);
+    negotiator.close();
   }
 
   @Test
@@ -162,7 +201,21 @@ public class NettyChannelBuilderTest {
     thrown.expect(NullPointerException.class);
     NettyChannelBuilder.createProtocolNegotiatorByType(
         NegotiationType.TLS,
-        noSslContext);
+        noSslContext, null);
+  }
+
+  @Test
+  public void createProtocolNegotiatorByType_tlsWithExecutor() throws Exception {
+    TrackingObjectPoolForTest executorPool = new TrackingObjectPoolForTest();
+    assertEquals(false, executorPool.isInUse());
+    SslContext localSslContext = GrpcSslContexts.forClient().build();
+    ProtocolNegotiator negotiator = NettyChannelBuilder.createProtocolNegotiatorByType(
+        NegotiationType.TLS,
+        localSslContext, executorPool);
+    assertEquals(true, executorPool.isInUse());
+    assertNotNull(negotiator);
+    negotiator.close();
+    assertEquals(false, executorPool.isInUse());
   }
 
   @Test
@@ -257,4 +310,18 @@ public class NettyChannelBuilderTest {
 
     builder.assertEventLoopAndChannelType();
   }
+
+  @Test
+  public void transportFactorySupportsNettyChannelCreds() {
+    NettyChannelBuilder builder = NettyChannelBuilder.forTarget("foo");
+    ClientTransportFactory transportFactory = builder.buildTransportFactory();
+
+    SwapChannelCredentialsResult result = transportFactory.swapChannelCredentials(
+        mock(ChannelCredentials.class));
+    assertThat(result).isNull();
+
+    result = transportFactory.swapChannelCredentials(
+        NettyChannelCredentials.create(new PlaintextProtocolNegotiatorClientFactory()));
+    assertThat(result).isNotNull();
+  }
 }
