@@ -22,6 +22,7 @@ import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertNull;
 
 import java.io.IOException;
+import java.net.SocketException;
 import java.util.logging.Level;
 import org.junit.Test;
 import org.junit.runner.RunWith;
@@ -34,6 +35,11 @@ public class NettyServerTransportTest {
     assertEquals(Level.INFO, getLogLevel(new Exception()));
   }
 
+  @Test
+  public void socketException() {
+    assertEquals(Level.FINE, getLogLevel(new SocketException("Connection reset")));
+  }
+
   @Test
   public void ioException() {
     assertEquals(Level.FINE, getLogLevel(new IOException("Connection reset by peer")));
