@@ -345,7 +345,7 @@ public class WriteBufferingAndExceptionHandlerTest {
     assertThat(chan.pipeline().context(handler)).isNull();
     assertThat(write.get().getClass()).isSameInstanceAs(Object.class);
     assertTrue(flush.get());
-    assertThat(chan.pipeline()).doesNotContain(handler);
+    assertThat(chan.pipeline().toMap().values()).doesNotContain(handler);
   }
 
   @Test
